from unittest.mock import Mock, patch

from django.test import RequestFactory, TestCase
from django.urls import reverse

from allianceauth.tests.auth_utils import AuthUtils
from esi.models import Token

from .. import views


class TestViews(TestCase):
        
    def setUp(self):
        self.factory = RequestFactory()
        self.user = AuthUtils.create_user('Bruce Wayne')        

    def test_groups_view_can_load(self):
        request = self.factory.get(reverse('groupmanagement:groups'))
        request.user = self.user
        response = views.groups_view(request)
        self.assertEqual(response.status_code, 200)        


from copy import deepcopy
from unittest import mock

from django.test import TestCase
from django.contrib.auth.models import Group, Permission

from allianceauth.authentication.models import State
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils


class ServicesSignalsTestCase(TestCase):
    def setUp(self):
        self.member = AuthUtils.create_user('auth_member', disconnect_signals=True)
        AuthUtils.add_main_character_2(
            self.member, 'Test', 1, 2, 'Test Corp', 'TEST'
        )
        self.none_user = AuthUtils.create_user('none_user', disconnect_signals=True)

    @mock.patch('allianceauth.services.signals.transaction')
    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_m2m_changed_user_groups(self, services_hook, transaction):
        """
        Test that update_groups hook function is called on user groups change
        """
        svc = mock.Mock()
        svc.update_groups.return_value = None
        svc.validate_user.return_value = None

        services_hook.get_services.return_value = [svc]

        # Overload transaction.on_commit so everything happens synchronously
        transaction.on_commit = lambda fn: fn()

        test_group = Group.objects.create(name="Test group")

        # Act, should trigger m2m change
        self.member.groups.add(test_group)
        self.member.save()

        # Assert
        self.assertTrue(services_hook.get_services.called)

        self.assertTrue(svc.update_groups.called)
        args, kwargs = svc.update_groups.call_args
        self.assertEqual(self.member, args[0])

        self.assertTrue(svc.validate_user.called)
        args, kwargs = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.disable_user')
    def test_pre_delete_user(self, disable_user):
        """
        Test that disable_member is called when a user is deleted
        """
        self.none_user.delete()

        self.assertTrue(disable_user.called)
        args, kwargs = disable_user.call_args
        self.assertEqual(self.none_user, args[0])

    @mock.patch('allianceauth.services.signals.disable_user')
    def test_pre_save_user_inactivation(self, disable_user):
        """
        Test a user set inactive has disable_member called
        """
        self.member.is_active = False
        self.member.save()  # Signal Trigger

        self.assertTrue(disable_user.called)
        args, kwargs = disable_user.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.disable_user')
    def test_disable_services_on_loss_of_main_character(self, disable_user):
        """
        Test a user set inactive has disable_member called
        """
        self.member.profile.main_character = None
        self.member.profile.save()  # Signal Trigger

        self.assertTrue(disable_user.called)
        args, kwargs = disable_user.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.transaction')
    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_m2m_changed_group_permissions(self, services_hook, transaction):
        from django.contrib.contenttypes.models import ContentType
        svc = mock.Mock()
        svc.validate_user.return_value = None
        svc.access_perm = 'auth.access_testsvc'

        services_hook.get_services.return_value = [svc]

        # Overload transaction.on_commit so everything happens synchronously
        transaction.on_commit = lambda fn: fn()

        test_group = Group.objects.create(name="Test group")
        AuthUtils.disconnect_signals()
        self.member.groups.add(test_group)
        AuthUtils.connect_signals()

        ct = ContentType.objects.get(app_label='auth', model='permission')
        perm = Permission.objects.create(name="Test perm", codename="access_testsvc", content_type=ct)
        test_group.permissions.add(perm)

        # Act, should trigger m2m change
        test_group.permissions.remove(perm)

        # Assert
        self.assertTrue(services_hook.get_services.called)

        self.assertTrue(svc.validate_user.called)
        args, kwargs = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.transaction')
    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_m2m_changed_user_permissions(self, services_hook, transaction):
        from django.contrib.contenttypes.models import ContentType
        svc = mock.Mock()
        svc.validate_user.return_value = None
        svc.access_perm = 'auth.access_testsvc'

        services_hook.get_services.return_value = [svc]

        # Overload transaction.on_commit so everything happens synchronously
        transaction.on_commit = lambda fn: fn()

        ct = ContentType.objects.get(app_label='auth', model='permission')
        perm = Permission.objects.create(
            name="Test perm", codename="access_testsvc", content_type=ct
        )
        self.member.user_permissions.add(perm)

        # Act, should trigger m2m change
        self.member.user_permissions.remove(perm)

        # Assert
        self.assertTrue(services_hook.get_services.called)

        self.assertTrue(svc.validate_user.called)
        args, kwargs = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.transaction')
    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_m2m_changed_user_state_permissions(self, services_hook, transaction):
        from django.contrib.contenttypes.models import ContentType
        svc = mock.Mock()
        svc.validate_user.return_value = None
        svc.access_perm = 'auth.access_testsvc'

        services_hook.get_services.return_value = [svc]

        # Overload transaction.on_commit so everything happens synchronously
        transaction.on_commit = lambda fn: fn()

        AuthUtils.disconnect_signals()
        test_state = State.objects.create(name="Test state", priority=150)
        self.member.profile.state = test_state
        self.member.profile.save()
        AuthUtils.connect_signals()

        ct = ContentType.objects.get(app_label='auth', model='permission')
        perm = Permission.objects.create(
            name="Test perm", codename="access_testsvc", content_type=ct
        )
        test_state.permissions.add(perm)

        # Act, should trigger m2m change
        test_state.permissions.remove(perm)

        # Assert
        self.assertTrue(services_hook.get_services.called)

        self.assertTrue(svc.validate_user.called)
        args, kwargs = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_state_changed_services_validation_and_groups_update(self, services_hook):
        """Test a user changing state has service accounts validated and groups updated
        """
        svc = mock.Mock()
        svc.validate_user.return_value = None
        svc.update_groups.return_value = None
        svc.access_perm = 'auth.access_testsvc'

        services_hook.get_services.return_value = [svc]

        test_state = State.objects.create(name="Test state", priority=150, public=True)
        self.member.profile.state = test_state
        self.member.profile.save()

        # Assert
        self.assertTrue(services_hook.get_services.called)

        self.assertTrue(svc.validate_user.called)        
        args, kwargs = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])
        
        self.assertTrue(svc.update_groups.called)
        args, kwargs = svc.update_groups.call_args
        self.assertEqual(self.member, args[0])


    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_state_changed_services_validation_and_groups_update_1(self, services_hook):
        """Test a user changing main has service accounts validated and sync updated
        """
        svc = mock.Mock()
        svc.validate_user.return_value = None
        svc.sync_nickname.return_value = None
        svc.access_perm = 'auth.access_testsvc'

        services_hook.get_services.return_value = [svc]
        
        new_main = EveCharacter.objects.create(
            character_id=123, 
            character_name="Alter Ego", 
            corporation_id=987, 
            corporation_name="ABC"
        )
        self.member.profile.main_character = new_main
        self.member.profile.save()

        # Assert
        self.assertTrue(services_hook.get_services.called)

        self.assertTrue(svc.validate_user.called)        
        args, kwargs = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])
        
        self.assertTrue(svc.sync_nickname.called)
        args, kwargs = svc.sync_nickname.call_args
        self.assertEqual(self.member, args[0])

    @mock.patch('allianceauth.services.signals.ServicesHook')
    def test_state_changed_services_validation_and_groups_update_2(self, services_hook):
        """Test a user changing main has service does not have accounts validated 
        and sync updated if the new main is equal to the old main
        """
        svc = mock.Mock()
        svc.validate_user.return_value = None
        svc.sync_nickname.return_value = None
        svc.access_perm = 'auth.access_testsvc'

        services_hook.get_services.return_value = [svc]
        
        # this creates a clone of the Django object
        new_main = deepcopy(self.member.profile.main_character)
        self.assertIsNot(new_main, self.member.profile.main_character)
        self.member.profile.main_character = new_main
        self.member.profile.save()

        # Assert
        self.assertFalse(services_hook.get_services.called)
        self.assertFalse(svc.validate_user.called)                
        self.assertFalse(svc.sync_nickname.called)


from django.contrib import admin

from .models import MumbleUser
from ...admin import ServicesUserAdmin


@admin.register(MumbleUser)
class MumbleUserAdmin(ServicesUserAdmin):            
    list_display = ServicesUserAdmin.list_display + (        
        'username',        
        'groups',
    )     
    search_fields = ServicesUserAdmin.search_fields + (        
        'username',
        'groups'
    )

    fields = ('user', 'username', 'groups')  # pwhash is hidden from admin panel



from django.contrib.auth.models import User
from django.db import models
from typing import Optional


class ApplicationManager(models.Manager):

    def pending_requests_count_for_user(self, user: User) -> Optional[int]:
        """Returns the number of pending group requests for the given user"""
        if user.is_superuser:
            return self.filter(approved__isnull=True).count()
        elif user.has_perm("auth.human_resources"):
            main_character = user.profile.main_character
            if main_character:
                return (
                    self
                    .select_related("form__corp")
                    .filter(form__corp__corporation_id=main_character.corporation_id)
                    .filter(approved__isnull=True)
                    .count()
                )
            else:
                return None
        else:
            return None



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-10 15:34
from __future__ import unicode_literals

from django.db import migrations


def delete_permissions(apps, schema_editor):
    CorpStats = apps.get_model('corputils', 'CorpStats')
    ContentType = apps.get_model('contenttypes', 'ContentType')
    Permission = apps.get_model('auth', 'Permission')
    ct = ContentType.objects.get_for_model(CorpStats)
    perms = Permission.objects.filter(content_type=ct)
    perms.filter(codename__contains='api').delete()
    perms.filter(codename='view_corpstats').delete()
    perms.filter(codename__contains='blue').delete()
    perms.filter(codename__contains='remove').delete()

    g = perms.get(codename='view_corp_corpstats')
    g.name = 'Can view corp stats of their corporation.'
    g.save()

    g = perms.get(codename='view_alliance_corpstats')
    g.name = 'Can view corp stats of members of their alliance.'
    g.save()


class Migration(migrations.Migration):

    dependencies = [
        ('corputils', '0004_member_models'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='corpstats',
            options={'permissions': (('view_corp_corpstats', 'Can view corp stats of their corporation.'), ('view_alliance_corpstats', 'Can view corp stats of members of their alliance.'), ('view_state_corpstats', 'Can view corp stats of members of their auth state.')), 'verbose_name': 'corp stats', 'verbose_name_plural': 'corp stats'},
        ),
        migrations.RunPython(delete_permissions, migrations.RunPython.noop),
    ]



import logging
from urllib.parse import urlencode

from requests_oauthlib import OAuth2Session
from requests.exceptions import HTTPError

from django.contrib.auth.models import User, Group
from django.db import models
from django.utils.timezone import now

from allianceauth.services.hooks import NameFormatter

from . import __title__
from .app_settings import (
    DISCORD_APP_ID,
    DISCORD_APP_SECRET,
    DISCORD_BOT_TOKEN,
    DISCORD_CALLBACK_URL,
    DISCORD_GUILD_ID,    
    DISCORD_SYNC_NAMES
)
from .discord_client import DiscordClient
from .discord_client.exceptions import DiscordClientException, DiscordApiBackoff
from .discord_client.helpers import match_or_create_roles_from_names
from .utils import LoggerAddTag


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class DiscordUserManager(models.Manager):
    """Manager for DiscordUser"""

    # full server admin
    BOT_PERMISSIONS = 0x00000008

    # get user ID, accept invite
    SCOPES = [
        'identify',
        'guilds.join',
    ]

    def add_user(
        self, 
        user: User, 
        authorization_code: str, 
        is_rate_limited: bool = True
    ) -> bool:
        """adds a new Discord user

        Params:
        - user: Auth user to join
        - authorization_code: authorization code returns from oauth
        - is_rate_limited: When False will disable default rate limiting (use with care)
        
        Returns: True on success, else False or raises exception
        """
        try:        
            nickname = self.user_formatted_nick(user) if DISCORD_SYNC_NAMES else None
            group_names = self.user_group_names(user)
            access_token = self._exchange_auth_code_for_token(authorization_code)
            user_client = DiscordClient(access_token, is_rate_limited=is_rate_limited)
            discord_user = user_client.current_user()
            user_id = discord_user['id']
            bot_client = self._bot_client(is_rate_limited=is_rate_limited)
                
            if group_names:                
                role_ids = match_or_create_roles_from_names(
                    client=bot_client, 
                    guild_id=DISCORD_GUILD_ID, 
                    role_names=group_names
                ).ids()
            else:
                role_ids = None
                        
            created = bot_client.add_guild_member(
                guild_id=DISCORD_GUILD_ID, 
                user_id=user_id,
                access_token=access_token,
                role_ids=role_ids,
                nick=nickname
            )
            if created is not False:
                if created is None:
                    logger.debug(
                        "User %s with Discord ID %s is already a member.",
                        user,
                        user_id,
                    )
                self.update_or_create(
                    user=user, 
                    defaults={
                        'uid': user_id, 
                        'username': discord_user['username'][:32], 
                        'discriminator': discord_user['discriminator'][:4],
                        'activated': now()
                    }
                )
                logger.info(
                    "Added user %s with Discord ID %s to Discord server", user, user_id
                )
                return True

            else:
                logger.warning(
                    "Failed to add user %s with Discord ID %s to Discord server",
                    user,
                    user_id,
                )
                return False

        except (HTTPError, ConnectionError, DiscordApiBackoff) as ex:
            logger.exception(
                'Failed to add user %s to Discord server: %s', user, ex
            )
            return False

    @staticmethod
    def user_formatted_nick(user: User) -> str:
        """returns the name of the given users main character with name formatting
        or None if user has no main
        """
        from .auth_hooks import DiscordService
        
        if user.profile.main_character:
            return NameFormatter(DiscordService(), user).format_name()
        else:
            return None

    @staticmethod
    def user_group_names(user: User, state_name: str = None) -> list:
        """returns list of group names plus state the given user is a member of"""
        if not state_name:
            state_name = user.profile.state.name
        group_names = (
            [group.name for group in user.groups.all()] + [state_name]
        )
        logger.debug(
            "Group names for roles updates of user %s are: %s", user, group_names
        )
        return group_names
    
    def user_has_account(self, user: User) -> bool:
        """Returns True if the user has an Discord account, else False
        
        only checks locally, does not hit the API
        """
        if not isinstance(user, User):
            return False
        return self.filter(user=user).select_related('user').exists()

    @classmethod
    def generate_bot_add_url(cls) -> str:
        params = urlencode({
            'client_id': DISCORD_APP_ID,
            'scope': 'bot',
            'permissions': str(cls.BOT_PERMISSIONS)

        })
        return f'{DiscordClient.OAUTH_BASE_URL}?{params}'        

    @classmethod
    def generate_oauth_redirect_url(cls) -> str:
        oauth = OAuth2Session(
            DISCORD_APP_ID, redirect_uri=DISCORD_CALLBACK_URL, scope=cls.SCOPES
        )
        url, state = oauth.authorization_url(DiscordClient.OAUTH_BASE_URL)
        return url

    @staticmethod
    def _exchange_auth_code_for_token(authorization_code: str) -> str:
        oauth = OAuth2Session(DISCORD_APP_ID, redirect_uri=DISCORD_CALLBACK_URL)
        token = oauth.fetch_token(
            DiscordClient.OAUTH_TOKEN_URL, 
            client_secret=DISCORD_APP_SECRET, 
            code=authorization_code
        )
        logger.debug("Received token from OAuth")
        return token['access_token']
    
    @classmethod
    def server_name(cls, use_cache: bool = True) -> str:
        """returns the name of the current Discord server 
        or an empty string if the name could not be retrieved

        Params:
        - use_cache: When set False will force an API call to get the server name
        """
        try:
            server_name = cls._bot_client().guild_name(
                guild_id=DISCORD_GUILD_ID, use_cache=use_cache
            )
        except (HTTPError, DiscordClientException):
            server_name = ""
        except Exception:
            logger.warning(
                "Unexpected error when trying to retrieve the server name from Discord", 
                exc_info=True
            )
            server_name = ""

        return server_name

    @classmethod
    def group_to_role(cls, group: Group) -> dict:
        """returns the Discord role matching the given Django group by name
        or an empty dict() if no matching role exist
        """
        return cls._bot_client().match_role_from_name(
            guild_id=DISCORD_GUILD_ID, role_name=group.name
        )

    @staticmethod
    def _bot_client(is_rate_limited: bool = True) -> DiscordClient:
        """returns a bot client for access to the Discord API"""
        return DiscordClient(DISCORD_BOT_TOKEN, is_rate_limited=is_rate_limited)



from django.conf.urls import url, include

from allianceauth.authentication import views

urlpatterns = [
    url(r'^activate/complete/$', views.activation_complete, name='registration_activation_complete'),
    # The activation key can make use of any character from the
    # URL-safe base64 alphabet, plus the colon as a separator.
    url(r'^activate/(?P<activation_key>[-:\w]+)/$', views.ActivationView.as_view(), name='registration_activate'),
    url(r'^register/$', views.RegistrationView.as_view(), name='registration_register'),
    url(r'^register/complete/$', views.registration_complete, name='registration_complete'),
    url(r'^register/closed/$', views.registration_closed, name='registration_disallowed'),
    url(r'', include('django.contrib.auth.urls')),
]


from unittest import mock
from urllib import parse

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.http.response import HttpResponse
from django.shortcuts import reverse
from django.test import TestCase
from django.test.client import RequestFactory

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

from ..decorators import main_character_required
from ..models import CharacterOwnership


MODULE_PATH = 'allianceauth.authentication'


class DecoratorTestCase(TestCase):
    @staticmethod
    @main_character_required
    def dummy_view(*args, **kwargs):
        return HttpResponse(status=200)

    @classmethod
    def setUpTestData(cls):
        cls.main_user = AuthUtils.create_user('main_user', disconnect_signals=True)
        cls.no_main_user = AuthUtils.create_user(
            'no_main_user', disconnect_signals=True
        )
        main_character = EveCharacter.objects.create(
            character_id=1,
            character_name='Main Character',
            corporation_id=1,
            corporation_name='Corp',
            corporation_ticker='CORP',
        )
        CharacterOwnership.objects.create(
            user=cls.main_user, character=main_character, owner_hash='1'
        )
        cls.main_user.profile.main_character = main_character

    def setUp(self):
        self.request = RequestFactory().get('/test/')

    @mock.patch(MODULE_PATH + '.decorators.messages')
    def test_login_redirect(self, m):
        setattr(self.request, 'user', AnonymousUser())
        response = self.dummy_view(self.request)
        self.assertEqual(response.status_code, 302)
        url = getattr(response, 'url', None)
        self.assertEqual(parse.urlparse(url).path, reverse(settings.LOGIN_URL))

    @mock.patch(MODULE_PATH + '.decorators.messages')
    def test_main_character_redirect(self, m):
        setattr(self.request, 'user', self.no_main_user)
        response = self.dummy_view(self.request)
        self.assertEqual(response.status_code, 302)
        url = getattr(response, 'url', None)
        self.assertEqual(url, reverse('authentication:dashboard'))

    @mock.patch(MODULE_PATH + '.decorators.messages')
    def test_successful_request(self, m):
        setattr(self.request, 'user', self.main_user)
        response = self.dummy_view(self.request)
        self.assertEqual(response.status_code, 200)



from allianceauth.authentication.signals import state_changed
from .managers import GroupManager
from .models import Group
from django.dispatch import receiver
import logging
logger = logging.getLogger(__name__)

@receiver(state_changed)
def check_groups_on_state_change(sender, user, state, **kwargs):
    logger.debug("Updating auth groups for {}".format(user))
    visible_groups = GroupManager.get_joinable_groups(state)
    visible_groups = visible_groups | Group.objects.select_related('authgroup').filter(authgroup__internal=True)
    groups = user.groups.all()
    for g in groups:
        if g not in visible_groups:
            user.groups.remove(g)



from unittest.mock import Mock, patch

from django.contrib.auth.models import Group, User
from django.test import TestCase
from django.urls import reverse

from allianceauth.eveonline.models import EveCorporationInfo, EveAllianceInfo
from allianceauth.tests.auth_utils import AuthUtils

from ..models import GroupRequest
from ..managers import GroupManager


class MockUserNotAuthenticated():
    def __init__(self):
        self.is_authenticated = False


class GroupManagementVisibilityTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('test')
        AuthUtils.add_main_character(
            cls.user, 'test character', '1', corp_id='2', corp_name='test_corp', corp_ticker='TEST', alliance_id='3', alliance_name='TEST'
        )
        cls.user.profile.refresh_from_db()
        cls.alliance = EveAllianceInfo.objects.create(alliance_id='3', alliance_name='test alliance', alliance_ticker='TEST', executor_corp_id='2')
        cls.corp = EveCorporationInfo.objects.create(
            corporation_id='2', corporation_name='test corp', corporation_ticker='TEST', alliance=cls.alliance, member_count=1
        )
        cls.group1 = Group.objects.create(name='group1')
        cls.group2 = Group.objects.create(name='group2')
        cls.group3 = Group.objects.create(name='group3')

    def setUp(self):
        self.user.refresh_from_db()

    def _refresh_user(self):
        self.user = User.objects.get(pk=self.user.pk)

    def test_get_group_leaders_groups(self):
        self.group1.authgroup.group_leaders.add(self.user)
        self.group2.authgroup.group_leader_groups.add(self.group1)
        self._refresh_user()
        groups = GroupManager.get_group_leaders_groups(self.user)

        self.assertIn(self.group1, groups)  #avail due to user
        self.assertNotIn(self.group2, groups)  #not avail due to group
        self.assertNotIn(self.group3, groups)  #not avail at all 

        self.user.groups.add(self.group1)
        self._refresh_user()
        groups = GroupManager.get_group_leaders_groups(self.user)

    def test_can_manage_group(self):
        self.group1.authgroup.group_leaders.add(self.user)
        self.user.groups.add(self.group1)
        self._refresh_user()

        self.assertTrue(GroupManager.can_manage_group(self.user, self.group1))
        self.assertFalse(GroupManager.can_manage_group(self.user, self.group2))
        self.assertFalse(GroupManager.can_manage_group(self.user, self.group3))
        
        self.group2.authgroup.group_leader_groups.add(self.group1)
        self.group1.authgroup.group_leaders.remove(self.user)
        self._refresh_user()

        self.assertFalse(GroupManager.can_manage_group(self.user, self.group1))
        self.assertTrue(GroupManager.can_manage_group(self.user, self.group2))
        self.assertFalse(GroupManager.can_manage_group(self.user, self.group3))


class TestGroupManager(TestCase):
        
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # group 1
        cls.group_default = Group.objects.create(name='default')        
        cls.group_default.authgroup.description = 'Default Group'
        cls.group_default.authgroup.internal = False
        cls.group_default.authgroup.hidden = False
        cls.group_default.authgroup.save()

        # group 2
        cls.group_internal = Group.objects.create(name='internal')
        cls.group_internal.authgroup.description = 'Internal Group'
        cls.group_internal.authgroup.internal = True        
        cls.group_internal.authgroup.save()

        # group 3
        cls.group_hidden = Group.objects.create(name='hidden')
        cls.group_hidden.authgroup.description = 'Hidden Group'
        cls.group_hidden.authgroup.internal = False
        cls.group_hidden.authgroup.hidden = True
        cls.group_hidden.authgroup.save()

        # group 4
        cls.group_open = Group.objects.create(name='open')
        cls.group_open.authgroup.description = 'Open Group'
        cls.group_open.authgroup.internal = False
        cls.group_open.authgroup.hidden = False
        cls.group_open.authgroup.open = True
        cls.group_open.authgroup.save()

        # group 5
        cls.group_public_1 = Group.objects.create(name='public 1')
        cls.group_public_1.authgroup.description = 'Public Group 1'
        cls.group_public_1.authgroup.internal = False
        cls.group_public_1.authgroup.hidden = False
        cls.group_public_1.authgroup.public = True
        cls.group_public_1.authgroup.save()

        # group 6
        cls.group_public_2 = Group.objects.create(name='public 2')
        cls.group_public_2.authgroup.description = 'Public Group 2'
        cls.group_public_2.authgroup.internal = False
        cls.group_public_2.authgroup.hidden = True
        cls.group_public_2.authgroup.open = True
        cls.group_public_2.authgroup.public = True
        cls.group_public_2.authgroup.save()

        # group 7
        cls.group_default_member = Group.objects.create(name='default members')
        cls.group_default_member.authgroup.description = \
            'Default Group for members only'
        cls.group_default_member.authgroup.internal = False
        cls.group_default_member.authgroup.hidden = False
        cls.group_default_member.authgroup.open = False
        cls.group_default_member.authgroup.public = False
        cls.group_default_member.authgroup.states.add(
            AuthUtils.get_member_state()
        )
        cls.group_default_member.authgroup.save()
    
    def setUp(self):
        self.user = AuthUtils.create_user('Bruce Wayne')

    def test_get_joinable_group_member(self):        
        result = GroupManager.get_joinable_groups(
            AuthUtils.get_member_state()
        )
        expected = {
            self.group_default, 
            self.group_hidden, 
            self.group_open, 
            self.group_public_1, 
            self.group_public_2,
            self.group_default_member
        }
        self.assertSetEqual(set(result), expected)

    def test_get_joinable_group_guest(self):
        result = GroupManager.get_joinable_groups(
            AuthUtils.get_guest_state()
        )
        expected = {
            self.group_default, 
            self.group_hidden, 
            self.group_open, 
            self.group_public_1, 
            self.group_public_2
        }
        self.assertSetEqual(set(result), expected)

    def test_joinable_group_member(self):
        member_state = AuthUtils.get_member_state()
        for x in [
            self.group_default, 
            self.group_hidden, 
            self.group_open, 
            self.group_public_1, 
            self.group_public_2,
            self.group_default_member
        ]:
            self.assertTrue(GroupManager.joinable_group(x, member_state))

        for x in [
            self.group_internal,             
        ]:
            self.assertFalse(GroupManager.joinable_group(x, member_state))

    def test_joinable_group_guest(self):
        guest_state = AuthUtils.get_guest_state()
        for x in [
            self.group_default, 
            self.group_hidden, 
            self.group_open, 
            self.group_public_1, 
            self.group_public_2
        ]:
            self.assertTrue(GroupManager.joinable_group(x, guest_state))

        for x in [
            self.group_internal,
            self.group_default_member
        ]:
            self.assertFalse(GroupManager.joinable_group(x, guest_state))

    def test_get_all_non_internal_groups(self):
        result = GroupManager.get_all_non_internal_groups()
        expected = {
            self.group_default, 
            self.group_hidden, 
            self.group_open, 
            self.group_public_1, 
            self.group_public_2, 
            self.group_default_member
        }
        self.assertSetEqual(set(result), expected)

    def test_check_internal_group(self):
        self.assertTrue(
            GroupManager.check_internal_group(self.group_default)
        )
        self.assertFalse(
            GroupManager.check_internal_group(self.group_internal)
        )

    def test_get_joinable_groups_for_user_no_permission(self):
        AuthUtils.assign_state(self.user, AuthUtils.get_guest_state())
        result = GroupManager.get_joinable_groups_for_user(self.user)
        expected = {self.group_public_1, self.group_public_2}
        self.assertSetEqual(set(result), expected)

    def test_get_joinable_groups_for_user_guest_w_permission_(self):
        AuthUtils.assign_state(self.user, AuthUtils.get_guest_state())
        AuthUtils.add_permission_to_user_by_name(
            'groupmanagement.request_groups', self.user
        )
        result = GroupManager.get_joinable_groups_for_user(self.user)
        expected = {
            self.group_default, 
            self.group_hidden, 
            self.group_open, 
            self.group_public_1, 
            self.group_public_2
        }
        self.assertSetEqual(set(result), expected)

    def test_get_joinable_groups_for_user_member_w_permission(self):
        AuthUtils.assign_state(self.user, AuthUtils.get_member_state(), True)
        AuthUtils.add_permission_to_user_by_name(
            'groupmanagement.request_groups', self.user
        )
        result = GroupManager.get_joinable_groups_for_user(self.user)
        expected = {
            self.group_default,
            self.group_hidden,
            self.group_open,
            self.group_public_1,
            self.group_public_2,
            self.group_default_member
        }
        self.assertSetEqual(set(result), expected)

    def test_get_joinable_groups_for_user_member_w_permission_no_hidden(self):
        AuthUtils.assign_state(self.user, AuthUtils.get_member_state(), True)
        AuthUtils.add_permission_to_user_by_name(
            'groupmanagement.request_groups', self.user
        )
        result = GroupManager.get_joinable_groups_for_user(
            self.user, include_hidden=False
        )
        expected = {
            self.group_default,
            self.group_open, 
            self.group_public_1,            
            self.group_default_member
        }
        self.assertSetEqual(set(result), expected)

    def test_has_management_permission(self):        
        user = AuthUtils.create_user('Clark Kent')
        AuthUtils.add_permission_to_user_by_name(
            'auth.group_management', user
        )        
        self.assertTrue(GroupManager.has_management_permission(user))

    def test_can_manage_groups_no_perm_no_group(self):        
        user = AuthUtils.create_user('Clark Kent')
        self.assertFalse(GroupManager.can_manage_groups(user))

    def test_can_manage_groups_user_not_authenticated(self):        
        user = MockUserNotAuthenticated()
        self.assertFalse(GroupManager.can_manage_groups(user))

    def test_can_manage_groups_has_perm(self):        
        user = AuthUtils.create_user('Clark Kent')
        AuthUtils.add_permission_to_user_by_name(
            'auth.group_management', user
        )        
        self.assertTrue(GroupManager.can_manage_groups(user))

    def test_can_manage_groups_no_perm_leads_group(self):        
        user = AuthUtils.create_user('Clark Kent')
        self.group_default.authgroup.group_leaders.add(user)
        self.assertTrue(GroupManager.can_manage_groups(user))

    def test_can_manage_group_no_perm_no_group(self):        
        user = AuthUtils.create_user('Clark Kent')
        self.assertFalse(
            GroupManager.can_manage_group(user, self.group_default)
        )

    def test_can_manage_group_has_perm(self):        
        user = AuthUtils.create_user('Clark Kent')
        AuthUtils.add_permission_to_user_by_name(
            'auth.group_management', user
        )        
        self.assertTrue(
            GroupManager.can_manage_group(user, self.group_default)
        )

    def test_can_manage_group_no_perm_leads_correct_group(self):        
        user = AuthUtils.create_user('Clark Kent')
        self.group_default.authgroup.group_leaders.add(user)
        self.assertTrue(
            GroupManager.can_manage_group(user, self.group_default)
        )

    def test_can_manage_group_no_perm_leads_other_group(self):        
        user = AuthUtils.create_user('Clark Kent')
        self.group_hidden.authgroup.group_leaders.add(user)
        self.assertFalse(
            GroupManager.can_manage_group(user, self.group_default)
        )

    def test_can_manage_group_user_not_authenticated(self):        
        user = MockUserNotAuthenticated()
        self.assertFalse(
            GroupManager.can_manage_group(user, self.group_default)
        )


class TestPendingRequestsCountForUser(TestCase):
            
    def setUp(self) -> None:
        self.group_1 = Group.objects.create(name="Group 1")
        self.group_2 = Group.objects.create(name="Group 2")
        self.user_leader_1 = AuthUtils.create_member('Clark Kent')
        self.group_1.authgroup.group_leaders.add(self.user_leader_1)
        self.user_leader_2 = AuthUtils.create_member('Peter Parker')
        self.group_2.authgroup.group_leaders.add(self.user_leader_2)
        self.user_requestor = AuthUtils.create_member('Bruce Wayne')

    def test_single_request_for_leader(self):        
        # given user_leader_1 is leader of group_1 
        # and user_leader_2 is leader of group_2
        # when user_requestor is requesting access to group 1
        # then return 1 for user_leader 1 and 0 for user_leader_2
        GroupRequest.objects.create(
            status="pending", user=self.user_requestor, group=self.group_1
        )
        self.assertEqual(
            GroupManager.pending_requests_count_for_user(self.user_leader_1), 1
        )
        self.assertEqual(
            GroupManager.pending_requests_count_for_user(self.user_leader_2), 0
        )
  
    def test_return_none_for_none_leader(self):
        # given user_requestor is leader of no group
        # when user_requestor is requesting access to group 1
        # then return 0 for user_requestor
        GroupRequest.objects.create(
            status="pending", user=self.user_requestor, group=self.group_1
        )
        self.assertEqual(
            GroupManager.pending_requests_count_for_user(self.user_requestor), 0
        )
   
    def test_single_leave_request(self):
        # given user_leader_2 is leader of group_2
        # and user_requestor is member of group 2
        # when user_requestor is requesting to leave group 2
        # then return 1 for user_leader_2
        self.user_requestor.groups.add(self.group_2)

        GroupRequest.objects.create(
            status="pending", 
            user=self.user_requestor, 
            group=self.group_2, 
            leave_request=True
        )
        self.assertEqual(
            GroupManager.pending_requests_count_for_user(self.user_leader_2), 1
        )

    def test_join_and_leave_request(self):
        # given user_leader_2 is leader of group_2
        # and user_requestor is member of group 2
        # when user_requestor is requesting to leave group 2 
        # and user_requestor_2 is requesting to join group 2
        # then return 2 for user_leader_2
        self.user_requestor.groups.add(self.group_2)
        user_requestor_2 = AuthUtils.create_member("Lex Luther")
        GroupRequest.objects.create(
            status="pending", 
            user=user_requestor_2, 
            group=self.group_2             
        )
        GroupRequest.objects.create(
            status="pending", 
            user=self.user_requestor, 
            group=self.group_2, 
            leave_request=True
        )
        self.assertEqual(
            GroupManager.pending_requests_count_for_user(self.user_leader_2), 2
        )

    def test_single_request_for_user_with_management_perm(self):        
        # given user_leader_4 which is leafer of no group
        # but has the management permissions
        # when user_requestor is requesting access to group 1
        # then return 1 for user_leader_4        
        user_leader_4 = AuthUtils.create_member("Lex Luther")
        AuthUtils.add_permission_to_user_by_name("auth.group_management", user_leader_4)
        user_leader_4 = User.objects.get(pk=user_leader_4.pk)
        GroupRequest.objects.create(
            status="pending", user=self.user_requestor, group=self.group_1
        )
        self.assertEqual(
            GroupManager.pending_requests_count_for_user(self.user_leader_1), 1
        )



from unittest import mock

from django.test import TestCase
from django.contrib.auth.models import User, Group

from allianceauth.eveonline.models import EveCorporationInfo, EveAllianceInfo
from allianceauth.tests.auth_utils import AuthUtils

from ..signals import check_groups_on_state_change


class GroupManagementStateTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('test')
        AuthUtils.add_main_character(
            cls.user, 'test character', '1', corp_id='2', corp_name='test_corp', corp_ticker='TEST', alliance_id='3', alliance_name='TEST'
        )
        cls.user.profile.refresh_from_db()
        cls.alliance = EveAllianceInfo.objects.create(
            alliance_id='3', alliance_name='test alliance', alliance_ticker='TEST', executor_corp_id='2'
        )
        cls.corp = EveCorporationInfo.objects.create(
            corporation_id='2', corporation_name='test corp', corporation_ticker='TEST', alliance=cls.alliance, member_count=1
        )
        cls.state_group = Group.objects.create(name='state_group')
        cls.open_group = Group.objects.create(name='open_group')
        cls.state = AuthUtils.create_state('test state', 500)
        cls.state_group.authgroup.states.add(cls.state)
        cls.state_group.authgroup.internal = False
        cls.state_group.save()

    def setUp(self):
        self.user.refresh_from_db()
        self.state.refresh_from_db()

    def _refresh_user(self):
        self.user = User.objects.get(pk=self.user.pk)

    def _refresh_test_group(self):
        self.state_group = Group.objects.get(pk=self.state_group.pk)

    def test_drop_state_group(self):
        self.user.groups.add(self.open_group)
        self.user.groups.add(self.state_group)     
        self.assertEqual(self.user.profile.state.name, "Guest")

        self.state.member_corporations.add(self.corp)
        self._refresh_user()
        self.assertEqual(self.user.profile.state, self.state)
        groups = self.user.groups.all()
        self.assertIn(self.state_group, groups) #keeps group
        self.assertIn(self.open_group, groups) #public group unafected

        self.state.member_corporations.clear()
        self._refresh_user()
        self.assertEqual(self.user.profile.state.name, "Guest")
        groups = self.user.groups.all()
        self.assertNotIn(self.state_group, groups) #looses group
        self.assertIn(self.open_group, groups) #public group unafected



import logging

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from celery import shared_task
from allianceauth.services.tasks import QueueOnce
from .models import MumbleUser

logger = logging.getLogger(__name__)


class MumbleTasks:
    def __init__(self):
        pass

    @staticmethod
    def has_account(user):
        try:
            return user.mumble.username != ''
        except ObjectDoesNotExist:
            return False

    @staticmethod
    def disable_mumble():
        logger.info("Deleting all MumbleUser models")
        MumbleUser.objects.all().delete()

    @staticmethod
    @shared_task(bind=True, name="mumble.update_groups", base=QueueOnce)
    def update_groups(self, pk):
        user = User.objects.get(pk=pk)
        logger.debug("Updating mumble groups for user %s" % user)
        if MumbleTasks.has_account(user):
            try:
                if not user.mumble.update_groups():
                    raise Exception("Group sync failed")
                logger.debug("Updated user %s mumble groups." % user)
                return True
            except MumbleUser.DoesNotExist:
                logger.info("Mumble group sync failed for {}, user does not have a mumble account".format(user))
            except:
                logger.exception("Mumble group sync failed for %s, retrying in 10 mins" % user)
                raise self.retry(countdown=60 * 10)
        else:
            logger.debug("User %s does not have a mumble account, skipping" % user)
        return False

    @staticmethod
    @shared_task(bind=True, name="mumble.update_display_name", base=QueueOnce)
    def update_display_name(self, pk):
        user = User.objects.get(pk=pk)
        logger.debug("Updating mumble groups for user %s" % user)
        if MumbleTasks.has_account(user):
            try:
                if not user.mumble.update_display_name():
                    raise Exception("Display Name Sync failed")
                logger.debug("Updated user %s mumble display name." % user)
                return True
            except MumbleUser.DoesNotExist:
                logger.info("Mumble display name sync failed for {}, user does not have a mumble account".format(user))
            except:
                logger.exception("Mumble display name sync failed for %s, retrying in 10 mins" % user)
                raise self.retry(countdown=60 * 10)
        else:
            logger.debug("User %s does not have a mumble account, skipping" % user)
        return False

    @staticmethod
    @shared_task(name="mumble.update_all_groups")
    def update_all_groups():
        logger.debug("Updating ALL mumble groups")
        for mumble_user in MumbleUser.objects.exclude(username__exact=''):
            MumbleTasks.update_groups.delay(mumble_user.user.pk)

    @staticmethod
    @shared_task(name="mumble.update_all_display_names")
    def update_all_display_names():
        logger.debug("Updating ALL mumble display names")
        for mumble_user in MumbleUser.objects.exclude(username__exact=''):
            MumbleTasks.update_display_name.delay(mumble_user.user.pk)




# Generated by Django 2.2.9 on 2020-03-16 07:49

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ('mumble', '0007_not_null_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='mumbleuser',
            name='display_name',
            field=models.CharField(max_length=254, null=True),
        )
    ]



import logging

from requests.exceptions import HTTPError

from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import gettext_lazy

from allianceauth.notifications import notify

from . import __title__
from .app_settings import DISCORD_GUILD_ID
from .discord_client import DiscordApiBackoff, DiscordRoles
from .discord_client.helpers import match_or_create_roles_from_names
from .managers import DiscordUserManager
from .utils import LoggerAddTag


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class DiscordUser(models.Model):

    USER_RELATED_NAME = 'discord'
    
    user = models.OneToOneField(
        User, 
        primary_key=True, 
        on_delete=models.CASCADE, 
        related_name=USER_RELATED_NAME,
        help_text='Auth user owning this Discord account'
    )
    uid = models.BigIntegerField(
        db_index=True,
        help_text='user\'s ID on Discord'
    )
    username = models.CharField(
        max_length=32, 
        default='', 
        blank=True,
        db_index=True,
        help_text='user\'s username on Discord'
    )
    discriminator = models.CharField(
        max_length=4, 
        default='', 
        blank=True, 
        help_text='user\'s discriminator on Discord'
    )
    activated = models.DateTimeField(
        default=None, 
        null=True, 
        blank=True,
        help_text='Date & time this service account was activated'
    )

    objects = DiscordUserManager()

    class Meta:
        permissions = (
            ("access_discord", "Can access the Discord service"),
        )

    def __str__(self):
        return f'{self.user.username} - {self.uid}'

    def __repr__(self):
        return f'{type(self).__name__}(user=\'{self.user}\', uid={self.uid})'

    def update_nickname(self, nickname: str = None) -> bool:
        """Update nickname with formatted name of main character

        Params:
        - nickname: optional nickname to be used instead of user's main

        Returns:
        - True on success
        - None if user is no longer a member of the Discord server
        - False on error or raises exception
        """        
        if not nickname:
            nickname = DiscordUser.objects.user_formatted_nick(self.user)
        if nickname:            
            client = DiscordUser.objects._bot_client()            
            success = client.modify_guild_member(
                guild_id=DISCORD_GUILD_ID,
                user_id=self.uid,
                nick=nickname
            )
            if success:
                logger.info('Nickname for %s has been updated', self.user)
            else:
                logger.warning('Failed to update nickname for %s', self.user)
            return success
         
        else:
            return False

    def update_groups(self, state_name: str = None) -> bool:
        """update groups for a user based on his current group memberships. 
        Will add or remove roles of a user as needed.
        
        Params:
        - state_name: optional state name to be used

        Returns:
        - True on success
        - None if user is no longer a member of the Discord server
        - False on error or raises exception
        """        
        client = DiscordUser.objects._bot_client()                                
        member_info = client.guild_member(guild_id=DISCORD_GUILD_ID, user_id=self.uid)
        if member_info is None:            
            # User is no longer a member
            return None
        
        guild_roles = DiscordRoles(client.guild_roles(guild_id=DISCORD_GUILD_ID))
        logger.debug('Current guild roles: %s', guild_roles.ids())
        if 'roles' in member_info:
            if not guild_roles.has_roles(member_info['roles']):
                guild_roles = DiscordRoles(
                    client.guild_roles(guild_id=DISCORD_GUILD_ID, use_cache=False)
                )
                if not guild_roles.has_roles(member_info['roles']):
                    raise RuntimeError(
                        'Member %s has unknown roles: %s' % (
                            self.user, 
                            set(member_info['roles']).difference(guild_roles.ids())
                        )
                    )
            member_roles = guild_roles.subset(member_info['roles'])
        else:
            raise RuntimeError('member_info from %s is not valid' % self.user)
                
        requested_roles = match_or_create_roles_from_names(
            client=client, 
            guild_id=DISCORD_GUILD_ID, 
            role_names=DiscordUser.objects.user_group_names(
                user=self.user, state_name=state_name
            )
        )
        logger.debug(
            'Requested roles for user %s: %s', self.user, requested_roles.ids()
        )
        logger.debug('Current roles user %s: %s', self.user, member_roles.ids())        
        member_roles_managed = member_roles.subset(managed_only=True)
        if requested_roles != member_roles.difference(member_roles_managed):
            logger.debug('Need to update roles for user %s', self.user)
            new_roles = requested_roles.union(member_roles_managed)
            success = client.modify_guild_member(
                guild_id=DISCORD_GUILD_ID,
                user_id=self.uid,
                role_ids=list(new_roles.ids())
            )
            if success:
                logger.info('Roles for %s have been updated', self.user)
            else:
                logger.warning('Failed to update roles for %s', self.user)
            return success

        else:
            logger.info('No need to update roles for user %s', self.user)
            return True

    def update_username(self) -> bool:
        """Updates the username incl. the discriminator 
        from the Discord server and saves it
                
        Returns:
        - True on success
        - None if user is no longer a member of the Discord server
        - False on error or raises exception
        """        
    
        client = DiscordUser.objects._bot_client()            
        user_info = client.guild_member(guild_id=DISCORD_GUILD_ID, user_id=self.uid)
        if user_info is None:
            success = None
        elif (
            user_info 
            and 'user' in user_info 
            and 'username' in user_info['user'] 
            and 'discriminator' in user_info['user']
        ):            
            self.username = user_info['user']['username']
            self.discriminator = user_info['user']['discriminator']
            self.save()
            logger.info('Username for %s has been updated', self.user)
            success = True
        else:
            logger.warning('Failed to update username for %s', self.user)
            success = False
        return success
                
    def delete_user(
        self, 
        notify_user: bool = False, 
        is_rate_limited: bool = True,
        handle_api_exceptions: bool = False
    ) -> bool:
        """Deletes the Discount user both on the server and locally

        Params:
        - notify_user: When True will sent a notification to the user 
        informing him about the deleting of his account
        - is_rate_limited: When False will disable default rate limiting (use with care)
        - handle_api_exceptions: When True method will return False 
        when an API exception occurs
        
        Returns True when successful, otherwise False or raises exceptions
        Return None if user does no longer exist
        """
        try:
            _user = self.user
            client = DiscordUser.objects._bot_client(is_rate_limited=is_rate_limited)
            success = client.remove_guild_member(
                guild_id=DISCORD_GUILD_ID, user_id=self.uid
            )
            if success is not False:
                deleted_count, _ = self.delete()
                if deleted_count > 0:
                    if notify_user:
                        notify(
                            user=_user, 
                            title=gettext_lazy('Discord Account Disabled'), 
                            message=gettext_lazy(
                                'Your Discord account was disabled automatically '
                                'by Auth. If you think this was a mistake, '
                                'please contact an admin.'
                            ),
                            level='warning'
                        )
                    logger.info('Account for user %s was deleted.', _user)
                    return True
                else:
                    logger.debug('Account for user %s was already deleted.', _user)
                    return None
            
            else:
                logger.warning(
                    'Failed to remove user %s from the Discord server', _user
                )
                return False
        
        except (HTTPError, ConnectionError, DiscordApiBackoff) as ex:
            if handle_api_exceptions:
                logger.exception(
                    'Failed to remove user %s from Discord server: %s',self.user, ex
                )
                return False        
            else:
                raise ex



from django_webtest import WebTest
from django import urls
from django.contrib.auth.models import Group, Permission

from allianceauth.tests.auth_utils import AuthUtils


class PermissionsToolViewsTestCase(WebTest):
    def setUp(self):
        self.member = AuthUtils.create_member('auth_member')
        AuthUtils.add_main_character(self.member, 'test character', '1234', '2345', 'test corp', 'testc')
        self.member.email = 'auth_member@example.com'
        self.member.save()
        self.none_user = AuthUtils.create_user('none_user', disconnect_signals=True)
        self.none_user2 = AuthUtils.create_user('none_user2', disconnect_signals=True)
        self.none_user3 = AuthUtils.create_user('none_user3', disconnect_signals=True)

        self.no_perm_user = AuthUtils.create_user('no_perm_user', disconnect_signals=True)

        AuthUtils.disconnect_signals()
        self.no_perm_group = Group.objects.create(name="No Permission Group")

        self.test_group = Group.objects.create(name="Test group")

        self.test_group.user_set.add(self.none_user)
        self.test_group.user_set.add(self.none_user2)
        self.test_group.user_set.add(self.none_user3)

        self.permission = Permission.objects.get_by_natural_key(codename='audit_permissions',
                                                                app_label='permissions_tool',
                                                                model='permissionstool')

        self.test_group.permissions.add(self.permission)
        self.member.user_permissions.add(self.permission)
        AuthUtils.connect_signals()

    def test_menu_item(self):
        self.app.set_user(self.member)
        response = self.app.get(urls.reverse('permissions_tool:overview'))

        response_content = response.content.decode('utf-8')

        self.assertInHTML('<li><a class="active" href="/permissions/overview/">'
                          '<i class="fas fa-id-card fa-fw"></i> Permissions Audit</a></li>', response_content)

    def test_permissions_overview(self):
        self.app.set_user(self.member)

        response = self.app.get(urls.reverse('permissions_tool:overview'))

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed('permissions_tool/overview.html')
        self.assertContains(response, self.permission.codename)
        self.assertContains(response, self.permission.content_type.app_label)
        self.assertContains(response, self.permission.content_type.model)

        tested_context = False
        # Test the context
        for perm in response.context['permissions']:
            if perm['permission'] == self.permission:
                tested_context = True
                self.assertIn('users', perm)
                self.assertEqual(perm['users'], 1)
                self.assertIn('groups', perm)
                self.assertEqual(perm['groups'], 1)
                self.assertIn('group_users', perm)
                self.assertEqual(perm['group_users'], 3)
                break
        self.assertTrue(tested_context)

    def test_permissions_overview_perms(self):
        # Ensure permission effectively denys access
        self.app.set_user(self.no_perm_user)

        response = self.app.get(urls.reverse('permissions_tool:overview'))

        self.assertEqual(response.status_code, 302)

    def test_permissions_audit(self):
        self.client.force_login(self.member)

        response = self.client.get(urls.reverse('permissions_tool:audit',
                                                kwargs={
                                                    'app_label': self.permission.content_type.app_label,
                                                    'model': self.permission.content_type.model,
                                                    'codename': self.permission.codename,
                                                }))

        self.assertTemplateUsed('permissions_tool/audit.html')
        self.assertTemplateUsed('permissions_tool/audit_row.html')

        self.assertContains(response, self.permission.codename)
        self.assertContains(response, self.none_user)
        self.assertContains(response, self.none_user3)
        self.assertContains(response, self.test_group)

        self.assertNotContains(response, self.no_perm_user)

    def test_permissions_audit_perms(self):
        # Ensure permission effectively denys access
        self.app.set_user(self.no_perm_user)

        response = self.app.get(urls.reverse('permissions_tool:audit',
                                                kwargs={
                                                    'app_label': self.permission.content_type.app_label,
                                                    'model': self.permission.content_type.model,
                                                    'codename': self.permission.codename,
                                                }))

        self.assertEqual(response.status_code, 302)



from unittest.mock import MagicMock, patch

from celery.exceptions import Retry
from requests.exceptions import HTTPError

from django.test import TestCase
from django.contrib.auth.models import Group
from django.test.utils import override_settings

from allianceauth.tests.auth_utils import AuthUtils

from . import TEST_USER_NAME, TEST_USER_ID, TEST_MAIN_NAME, TEST_MAIN_ID
from ..models import DiscordUser
from ..discord_client import DiscordApiBackoff
from .. import tasks
from ..utils import set_logger_to_file


MODULE_PATH = 'allianceauth.services.modules.discord.tasks'
logger = set_logger_to_file(MODULE_PATH, __file__)


@patch(MODULE_PATH + '.DiscordUser.update_groups')
@patch(MODULE_PATH + ".logger")
class TestUpdateGroups(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member(TEST_USER_NAME)
        cls.group_1 = Group.objects.create(name='Group 1')
        cls.group_2 = Group.objects.create(name='Group 2')
        cls.group_1.user_set.add(cls.user)
        cls.group_2.user_set.add(cls.user)
        
    def test_can_update_groups(self, mock_logger, mock_update_groups):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        tasks.update_groups(self.user.pk)
        self.assertTrue(mock_update_groups.called)

    def test_no_action_if_user_has_no_discord_account(
        self, mock_logger, mock_update_groups
    ):
        tasks.update_groups(self.user.pk)
        self.assertFalse(mock_update_groups.called)
    
    def test_retries_on_api_backoff(self, mock_logger, mock_update_groups):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        mock_exception = DiscordApiBackoff(999)
        mock_update_groups.side_effect = mock_exception

        with self.assertRaises(Retry):
            tasks.update_groups(self.user.pk)

    def test_retry_on_http_error_except_404(self, mock_logger, mock_update_groups):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        mock_exception = HTTPError('error')
        mock_exception.response = MagicMock()
        mock_exception.response.status_code = 500
        mock_update_groups.side_effect = mock_exception

        with self.assertRaises(Retry):
            tasks.update_groups(self.user.pk)

        self.assertTrue(mock_logger.warning.called)
    
    def test_retry_on_http_error_404_when_user_not_deleted(
        self, mock_logger, mock_update_groups
    ):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        mock_exception = HTTPError('error')
        mock_exception.response = MagicMock()
        mock_exception.response.status_code = 404
        mock_update_groups.side_effect = mock_exception

        with self.assertRaises(Retry):
            tasks.update_groups(self.user.pk)

        self.assertTrue(mock_logger.warning.called)
        
    def test_retry_on_non_http_error(self, mock_logger, mock_update_groups):        
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        mock_update_groups.side_effect = ConnectionError

        with self.assertRaises(Retry):
            tasks.update_groups(self.user.pk)

        self.assertTrue(mock_logger.warning.called)
        
    @patch(MODULE_PATH + '.DISCORD_TASKS_MAX_RETRIES', 3)
    def test_log_error_if_retries_exhausted(self, mock_logger, mock_update_groups):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)        
        mock_task = MagicMock(**{'request.retries': 3})
        mock_update_groups.side_effect = ConnectionError
        update_groups_inner = tasks.update_groups.__wrapped__.__func__
                
        update_groups_inner(mock_task, self.user.pk)
        self.assertTrue(mock_logger.error.called)
                
    @patch(MODULE_PATH + '.delete_user.delay')
    def test_delete_user_if_user_is_no_longer_member_of_discord_server(
        self, mock_delete_user, mock_logger, mock_update_groups
    ):
        mock_update_groups.return_value = None
        
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        tasks.update_groups(self.user.pk)
        self.assertTrue(mock_update_groups.called)
        self.assertTrue(mock_delete_user.called)


@patch(MODULE_PATH + '.DiscordUser.update_nickname')
class TestUpdateNickname(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member(TEST_USER_NAME)
        AuthUtils.add_main_character_2(
            cls.user, 
            TEST_MAIN_NAME, 
            TEST_MAIN_ID, 
            corp_id='2', 
            corp_name='test_corp', 
            corp_ticker='TEST',
            disconnect_signals=True
        )
        cls.discord_user = DiscordUser.objects.create(user=cls.user, uid=TEST_USER_ID)

    def test_can_update_nickname(self, mock_update_nickname):        
        mock_update_nickname.return_value = True        
        
        tasks.update_nickname(self.user.pk)
        self.assertTrue(mock_update_nickname.called)

    def test_no_action_when_user_had_no_account(self, mock_update_nickname):
        my_user = AuthUtils.create_user('Dummy User')
        mock_update_nickname.return_value = False
        
        tasks.update_nickname(my_user.pk)
        self.assertFalse(mock_update_nickname.called)
        
    def test_retries_on_api_backoff(self, mock_update_nickname):                
        mock_exception = DiscordApiBackoff(999)
        mock_update_nickname.side_effect = mock_exception
        
        with self.assertRaises(Retry):
            tasks.update_nickname(self.user.pk)

    def test_retries_on_general_exception(self, mock_update_nickname):        
        mock_update_nickname.side_effect = ConnectionError
        
        with self.assertRaises(Retry):
            tasks.update_nickname(self.user.pk)

    @patch(MODULE_PATH + '.DISCORD_TASKS_MAX_RETRIES', 3)
    def test_log_error_if_retries_exhausted(self, mock_update_nickname):
        mock_task = MagicMock(**{'request.retries': 3})
        mock_update_nickname.side_effect = ConnectionError
        update_nickname_inner = tasks.update_nickname.__wrapped__.__func__
                
        update_nickname_inner(mock_task, self.user.pk)
        

@patch(MODULE_PATH + '.DiscordUser.update_username')
class TestUpdateUsername(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member(TEST_USER_NAME)
        cls.discord_user = DiscordUser.objects.create(user=cls.user, uid=TEST_USER_ID)

    def test_can_update_username(self, mock_update_username):
        mock_update_username.return_value = True

        tasks.update_username(self.user.pk)
        self.assertTrue(mock_update_username.called)


@patch(MODULE_PATH + '.DiscordUser.delete_user')
class TestDeleteUser(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member(TEST_USER_NAME)
        cls.discord_user = DiscordUser.objects.create(user=cls.user, uid=TEST_USER_ID)

    def test_can_delete_user(self, mock_delete_user):
        mock_delete_user.return_value = True

        tasks.delete_user(self.user.pk)
        self.assertTrue(mock_delete_user.called)

    def test_can_delete_user_with_notify(self, mock_delete_user):
        mock_delete_user.return_value = True

        tasks.delete_user(self.user.pk, notify_user=True)
        self.assertTrue(mock_delete_user.called)
        args, kwargs = mock_delete_user.call_args
        self.assertTrue(kwargs['notify_user'])

    @patch(MODULE_PATH + '.delete_user.delay')
    def test_dont_retry_delete_user_if_user_is_no_longer_member_of_discord_server(
        self, mock_delete_user_delay, mock_delete_user
    ):
        mock_delete_user.return_value = None
        
        tasks.delete_user(self.user.pk)
        self.assertTrue(mock_delete_user.called)
        self.assertFalse(mock_delete_user_delay.called)


@patch(MODULE_PATH + '.DiscordUser.update_groups')
class TestTaskPerformUserAction(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member('Peter Parker')    
        cls.discord_user = DiscordUser.objects.create(user=cls.user, uid=TEST_USER_ID)

    def test_raise_value_error_on_unknown_method(self, mock_update_groups):
        mock_task = MagicMock(**{'request.retries': 0})
        
        with self.assertRaises(ValueError):
            tasks._task_perform_user_action(mock_task, self.user.pk, 'invalid_method')
    
    def test_catch_and_log_unexpected_exceptions(self, mock_update_groups):
        mock_task = MagicMock(**{'request.retries': 0})
        mock_update_groups.side_effect = RuntimeError

        tasks._task_perform_user_action(mock_task, self.user.pk, 'update_groups')


@patch(MODULE_PATH + '.DiscordUser.objects.server_name')
@patch(MODULE_PATH + ".logger")
class TestTaskUpdateServername(TestCase):

    def test_normal(self, mock_logger, mock_server_name):
        tasks.update_servername()
        self.assertTrue(mock_server_name.called)
        self.assertFalse(mock_logger.error.called)
        _, kwargs = mock_server_name.call_args
        self.assertFalse(kwargs["use_cache"])

    def test_retries_on_api_backoff(self, mock_logger, mock_server_name):
        mock_server_name.side_effect = DiscordApiBackoff(999)

        with self.assertRaises(Retry):
            tasks.update_servername()

        self.assertFalse(mock_logger.error.called)

    def test_retry_on_http_error(self, mock_logger, mock_server_name):        
        mock_exception = HTTPError(MagicMock(**{"response.status_code": 500}))        
        mock_server_name.side_effect = mock_exception

        with self.assertRaises(Retry):
            tasks.update_servername()

        self.assertTrue(mock_logger.warning.called)

    def test_retry_on_connection_error(self, mock_logger, mock_server_name):        
        mock_server_name.side_effect = ConnectionError

        with self.assertRaises(Retry):
            tasks.update_servername()

        self.assertTrue(mock_logger.warning.called)
    
    @patch(MODULE_PATH + '.DISCORD_TASKS_MAX_RETRIES', 3)
    def test_log_error_if_retries_exhausted(self, mock_logger, mock_server_name):
        mock_task = MagicMock(**{'request.retries': 3})
        mock_server_name.side_effect = ConnectionError
        update_groups_inner = tasks.update_servername.__wrapped__.__func__
                
        update_groups_inner(mock_task)
        self.assertTrue(mock_logger.error.called)


@patch(MODULE_PATH + '.DiscordUser.objects.server_name')
class TestTaskPerformUsersAction(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()        

    def test_raise_value_error_on_unknown_method(self, mock_server_name):
        mock_task = MagicMock(**{'request.retries': 0})
        
        with self.assertRaises(ValueError):
            tasks._task_perform_users_action(mock_task, 'invalid_method')
    
    def test_catch_and_log_unexpected_exceptions(self, mock_server_name):
        mock_server_name.side_effect = RuntimeError
        mock_task = MagicMock(**{'request.retries': 0})
        
        tasks._task_perform_users_action(mock_task, 'server_name')


@override_settings(CELERY_ALWAYS_EAGER=True)
class TestBulkTasks(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_1 = AuthUtils.create_user('Peter Parker')
        cls.user_2 = AuthUtils.create_user('Kara Danvers')
        cls.user_3 = AuthUtils.create_user('Clark Kent')
        DiscordUser.objects.all().delete()

    @patch(MODULE_PATH + '.update_groups.si')
    def test_can_update_groups_for_multiple_users(self, mock_update_groups):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        DiscordUser.objects.create(user=self.user_3, uid=789)
        expected_pks = [du_1.pk, du_2.pk]

        tasks.update_groups_bulk(expected_pks)
        self.assertEqual(mock_update_groups.call_count, 2)
        current_pks = [args[0][0] for args in mock_update_groups.call_args_list]
        
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.update_groups.si')
    def test_can_update_all_groups(self, mock_update_groups):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        du_3 = DiscordUser.objects.create(user=self.user_3, uid=789)

        tasks.update_all_groups()
        self.assertEqual(mock_update_groups.call_count, 3)
        current_pks = [args[0][0] for args in mock_update_groups.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.update_nickname.si')
    def test_can_update_nicknames_for_multiple_users(self, mock_update_nickname):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        DiscordUser.objects.create(user=self.user_3, uid=789)
        expected_pks = [du_1.pk, du_2.pk]

        tasks.update_nicknames_bulk(expected_pks)
        self.assertEqual(mock_update_nickname.call_count, 2)
        current_pks = [
            args[0][0] for args in mock_update_nickname.call_args_list
        ]        
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.update_nickname.si')
    def test_can_update_nicknames_for_all_users(self, mock_update_nickname):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid='123')
        du_2 = DiscordUser.objects.create(user=self.user_2, uid='456')
        du_3 = DiscordUser.objects.create(user=self.user_3, uid='789')

        tasks.update_all_nicknames()
        self.assertEqual(mock_update_nickname.call_count, 3)
        current_pks = [
            args[0][0] for args in mock_update_nickname.call_args_list
        ]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.update_username.si')
    def test_can_update_username_for_multiple_users(self, mock_update_username):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        DiscordUser.objects.create(user=self.user_3, uid=789)
        expected_pks = [du_1.pk, du_2.pk]

        tasks.update_usernames_bulk(expected_pks)
        self.assertEqual(mock_update_username.call_count, 2)
        current_pks = [args[0][0] for args in mock_update_username.call_args_list]
        
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.update_username')
    @patch(MODULE_PATH + '.update_servername')    
    def test_can_update_all_usernames(
        self, mock_update_servername, mock_update_username
    ):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        du_3 = DiscordUser.objects.create(user=self.user_3, uid=789)

        tasks.update_all_usernames()
        self.assertTrue(mock_update_servername.delay.called)
        self.assertEqual(mock_update_username.si.call_count, 3)
        current_pks = [args[0][0] for args in mock_update_username.si.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.DISCORD_SYNC_NAMES', True)
    @patch(MODULE_PATH + '.update_nickname')
    @patch(MODULE_PATH + '.update_groups')
    @patch(MODULE_PATH + '.update_username')
    def test_can_update_all_incl_nicknames(
        self, mock_update_usernames, mock_update_groups, mock_update_nickname
    ):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        du_3 = DiscordUser.objects.create(user=self.user_3, uid=789)

        tasks.update_all()
        self.assertEqual(mock_update_groups.si.call_count, 3)        
        current_pks = [args[0][0] for args in mock_update_groups.si.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

        self.assertEqual(mock_update_nickname.si.call_count, 3)
        current_pks = [args[0][0] for args in mock_update_nickname.si.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

        self.assertEqual(mock_update_usernames.si.call_count, 3)        
        current_pks = [args[0][0] for args in mock_update_usernames.si.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

    @patch(MODULE_PATH + '.DISCORD_SYNC_NAMES', False)
    @patch(MODULE_PATH + '.update_nickname')
    @patch(MODULE_PATH + '.update_groups')
    @patch(MODULE_PATH + '.update_username')
    def test_can_update_all_excl_nicknames(
        self, mock_update_usernames, mock_update_groups, mock_update_nickname
    ):
        du_1 = DiscordUser.objects.create(user=self.user_1, uid=123)
        du_2 = DiscordUser.objects.create(user=self.user_2, uid=456)
        du_3 = DiscordUser.objects.create(user=self.user_3, uid=789)

        tasks.update_all()
        self.assertEqual(mock_update_groups.si.call_count, 3)        
        current_pks = [args[0][0] for args in mock_update_groups.si.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))

        self.assertEqual(mock_update_nickname.si.call_count, 0)

        self.assertEqual(mock_update_usernames.si.call_count, 3)        
        current_pks = [args[0][0] for args in mock_update_usernames.si.call_args_list]
        expected_pks = [du_1.pk, du_2.pk, du_3.pk]
        self.assertSetEqual(set(current_pks), set(expected_pks))



from django.contrib import admin

from .models import OpenfireUser
from ...admin import ServicesUserAdmin


@admin.register(OpenfireUser)
class OpenfireUserAdmin(ServicesUserAdmin):
    list_display = ServicesUserAdmin.list_display + ('username',)         
    search_fields = ServicesUserAdmin.search_fields + ('username', )




from pydiscourse import DiscourseClient
from django.conf import settings

class DiscourseAPIClient():
    _client = None

    def __init__(self):
        pass

    @property
    def client(self):
        if not self._client:
            self._client = DiscourseClient(
                                settings.DISCOURSE_URL,
                                api_username=settings.DISCOURSE_API_USERNAME,
                                api_key=settings.DISCOURSE_API_KEY)
        return self._client

discourse = DiscourseAPIClient()


from django.conf import settings
from .views import NightModeRedirectView


def auth_settings(request):
    return {
        'SITE_NAME': settings.SITE_NAME,
        'NIGHT_MODE': NightModeRedirectView.night_mode_state(request),
    }



from django.utils.translation import ugettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from . import urls
from .models import Application


class ApplicationsMenu(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(self,
                              _('Applications'),
                              'far fa-file fa-fw',
                              'hrapplications:index',
                              navactive=['hrapplications:'])

    def render(self, request):        
        app_count = Application.objects.pending_requests_count_for_user(request.user)
        self.count = app_count if app_count and app_count > 0 else None
        return MenuItemHook.render(self, request)
    

@hooks.register('menu_item_hook')
def register_menu():
    return ApplicationsMenu()


class ApplicationsUrls(UrlHook):
    def __init__(self):
        UrlHook.__init__(self, urls, 'hrapplications', r'^hr/')


@hooks.register('url_hook')
def register_url():
    return ApplicationsUrls()



from django.contrib.auth.models import User, Group
from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

from esi.models import Token

from ..backends import StateBackend
from ..models import CharacterOwnership, UserProfile, OwnershipRecord

MODULE_PATH = 'allianceauth.authentication'

PERMISSION_1 = "authentication.add_user"
PERMISSION_2 = "authentication.change_user"


class TestStatePermissions(TestCase):

    def setUp(self):
        # permissions
        self.permission_1 = AuthUtils.get_permission_by_name(PERMISSION_1)
        self.permission_2 = AuthUtils.get_permission_by_name(PERMISSION_2)

        # group
        self.group_1 = Group.objects.create(name="Group 1")        
        self.group_2 = Group.objects.create(name="Group 2")

        # state
        self.state_1 = AuthUtils.get_member_state()        
        self.state_2 = AuthUtils.create_state("Other State", 75)
        
        # user
        self.user = AuthUtils.create_user("Bruce Wayne")
        self.main = AuthUtils.add_main_character_2(self.user, self.user.username, 123)

    def test_user_has_user_permissions(self):
        self.user.user_permissions.add(self.permission_1)
        
        user = User.objects.get(pk=self.user.pk)
        self.assertTrue(user.has_perm(PERMISSION_1))

    def test_user_has_group_permissions(self):        
        self.group_1.permissions.add(self.permission_1)
        self.user.groups.add(self.group_1)
        
        user = User.objects.get(pk=self.user.pk)
        self.assertTrue(user.has_perm(PERMISSION_1))

    def test_user_has_state_permissions(self):
        self.state_1.permissions.add(self.permission_1)
        self.state_1.member_characters.add(self.main)
        user = User.objects.get(pk=self.user.pk)

        self.assertTrue(user.has_perm(PERMISSION_1))

    def test_when_user_changes_state_perms_change_accordingly(self):
        self.state_1.permissions.add(self.permission_1)        
        self.state_1.member_characters.add(self.main)
        user = User.objects.get(pk=self.user.pk)
        self.assertTrue(user.has_perm(PERMISSION_1))

        self.state_2.permissions.add(self.permission_2)
        self.state_2.member_characters.add(self.main)
        self.state_1.member_characters.remove(self.main)
        user = User.objects.get(pk=self.user.pk)
        self.assertFalse(user.has_perm(PERMISSION_1))
        self.assertTrue(user.has_perm(PERMISSION_2))

    def test_state_permissions_are_returned_for_current_user_object(self):
        # verify state permissions are returns for the current user object 
        # and not for it's instance in the database, which might be outdated        
        self.state_1.permissions.add(self.permission_1)
        self.state_2.permissions.add(self.permission_2)        
        self.state_1.member_characters.add(self.main)
        user = User.objects.get(pk=self.user.pk)
        user.profile.state = self.state_2
        self.assertFalse(user.has_perm(PERMISSION_1))
        self.assertTrue(user.has_perm(PERMISSION_2))
    

class TestAuthenticate(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.main_character = EveCharacter.objects.create(
            character_id=1,
            character_name='Main Character',
            corporation_id=1,
            corporation_name='Corp',
            corporation_ticker='CORP',
        )
        cls.alt_character = EveCharacter.objects.create(
            character_id=2,
            character_name='Alt Character',
            corporation_id=1,
            corporation_name='Corp',
            corporation_ticker='CORP',
        )
        cls.unclaimed_character = EveCharacter.objects.create(
            character_id=3,
            character_name='Unclaimed Character',
            corporation_id=1,
            corporation_name='Corp',
            corporation_ticker='CORP',
        )
        cls.user = AuthUtils.create_user('test_user', disconnect_signals=True)
        cls.old_user = AuthUtils.create_user('old_user', disconnect_signals=True)
        AuthUtils.disconnect_signals()
        CharacterOwnership.objects.create(user=cls.user, character=cls.main_character, owner_hash='1')
        CharacterOwnership.objects.create(user=cls.user, character=cls.alt_character, owner_hash='2')
        UserProfile.objects.update_or_create(user=cls.user, defaults={'main_character': cls.main_character})
        AuthUtils.connect_signals()

    def test_authenticate_main_character(self):
        t = Token(character_id=self.main_character.character_id, character_owner_hash='1')
        user = StateBackend().authenticate(token=t)
        self.assertEquals(user, self.user)

    def test_authenticate_alt_character(self):
        t = Token(character_id=self.alt_character.character_id, character_owner_hash='2')
        user = StateBackend().authenticate(token=t)
        self.assertEquals(user, self.user)

    def test_authenticate_unclaimed_character(self):
        t = Token(character_id=self.unclaimed_character.character_id, character_name=self.unclaimed_character.character_name, character_owner_hash='3')
        user = StateBackend().authenticate(token=t)
        self.assertNotEqual(user, self.user)
        self.assertEqual(user.username, 'Unclaimed_Character')
        self.assertEqual(user.profile.main_character, self.unclaimed_character)

    def test_authenticate_character_record(self):
        t = Token(character_id=self.unclaimed_character.character_id, character_name=self.unclaimed_character.character_name, character_owner_hash='4')
        OwnershipRecord.objects.create(user=self.old_user, character=self.unclaimed_character, owner_hash='4')
        user = StateBackend().authenticate(token=t)
        self.assertEqual(user, self.old_user)
        self.assertTrue(CharacterOwnership.objects.filter(owner_hash='4', user=self.old_user).exists())
        self.assertTrue(user.profile.main_character)

    def test_iterate_username(self):
        t = Token(character_id=self.unclaimed_character.character_id,
                  character_name=self.unclaimed_character.character_name, character_owner_hash='3')
        username = StateBackend().authenticate(token=t).username
        t.character_owner_hash = '4'
        username_1 = StateBackend().authenticate(token=t).username
        t.character_owner_hash = '5'
        username_2 = StateBackend().authenticate(token=t).username
        self.assertNotEqual(username, username_1, username_2)
        self.assertTrue(username_1.endswith('_1'))
        self.assertTrue(username_2.endswith('_2'))



from django.test import TestCase
from django.contrib.auth.models import Group
from django.db import transaction

from allianceauth.tests.auth_utils import AuthUtils

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo, EveAllianceInfo

from ..models import AutogroupsConfig, get_users_for_state


from . import patch, connect_signals, disconnect_signals


class AutogroupsConfigTestCase(TestCase):
    def setUp(self):
        # Disconnect signals
        disconnect_signals()

        state = AuthUtils.get_member_state()

        self.alliance = EveAllianceInfo.objects.create(
            alliance_id='3456',
            alliance_name='alliance name',
            alliance_ticker='TIKR',
            executor_corp_id='2345',
        )

        self.corp = EveCorporationInfo.objects.create(
            corporation_id='2345',
            corporation_name='corp name',
            corporation_ticker='TIKK',
            member_count=10,
            alliance=self.alliance,
        )

        state.member_alliances.add(self.alliance)
        state.member_corporations.add(self.corp)

        self.member = AuthUtils.create_member('test user')

    def tearDown(self):
        # Reconnect signals
        connect_signals()

    def test_get_users_for_state(self):
        result = get_users_for_state(self.member.profile.state)

        self.assertIn(self.member, result)
        self.assertEqual(len(result), 1)

    @patch('.models.AutogroupsConfig.update_alliance_group_membership')
    @patch('.models.AutogroupsConfig.update_corp_group_membership')
    def test_update_group_membership_for_user(
        self, 
        update_corp, 
        update_alliance
    ):
        agc = AutogroupsConfig.objects.create()
        agc.update_group_membership_for_user(self.member)

        self.assertTrue(update_corp.called)
        self.assertTrue(update_alliance.called)

        args, kwargs = update_corp.call_args
        self.assertEqual(args[0], self.member)

        args, kwargs = update_alliance.call_args
        self.assertEqual(args[0], self.member)

    def test_update_alliance_group_membership(self):
        obj = AutogroupsConfig.objects.create(alliance_groups=True)
        obj.states.add(AuthUtils.get_member_state())
        char = EveCharacter.objects.create(
            character_id='1234',
            character_name='test character',
            corporation_id='2345',
            corporation_name='test corp',
            corporation_ticker='tickr',
            alliance_id='3456',
            alliance_name='alliance name',
        )
        self.member.profile.main_character = char
        self.member.profile.save()

        pre_groups = self.member.groups.all()

        # Act
        obj.update_alliance_group_membership(self.member)
        obj.update_corp_group_membership(self.member)  # check for no side effects

        group = obj.create_alliance_group(self.alliance)
        group_qs = Group.objects.filter(pk=group.pk)

        self.assertIn(group, self.member.groups.all())
        self.assertQuerysetEqual(self.member.groups.all(), map(repr, pre_groups | group_qs), ordered=False)

    def test_update_alliance_group_membership_no_main_character(self):
        obj = AutogroupsConfig.objects.create()
        obj.states.add(AuthUtils.get_member_state())

        # Act
        obj.update_alliance_group_membership(self.member)

        group = obj.get_alliance_group(self.alliance)

        self.assertNotIn(group, self.member.groups.all())

    # todo: this test case currently does not work, because it forces
    # an exception during a transaction, which is not easily testable
    # the production code itself should be fine though
    # I therefore commented out the test case for now
    """
    @patch('.models.EveAllianceInfo.objects.create_alliance')
    def test_update_alliance_group_membership_no_alliance_model(
        self,
        mock_create_alliance
    ):
        def mock_create_alliance_side_effect(*args, **kwargs):
            return EveAllianceInfo.objects.create(
                alliance_id='3459',
                alliance_name='alliance name',
                alliance_ticker='alliance_ticker',
                executor_corp_id='2345'
            )
        
        mock_create_alliance.side_effect = mock_create_alliance_side_effect
        
        obj = AutogroupsConfig.objects.create(alliance_groups=True)
        obj.states.add(AuthUtils.get_member_state())
        char = EveCharacter.objects.create(
            character_id='1234',
            character_name='test character',
            corporation_id='2345',
            corporation_name='test corp',
            corporation_ticker='tickr',
            alliance_id='3459',
            alliance_name='alliance name',
        )
        self.member.profile.main_character = char
        self.member.profile.save()

        # Act        
        obj.update_alliance_group_membership(self.member)

        group = obj.get_alliance_group(self.alliance)

        self.assertNotIn(group, self.member.groups.all())
    """

    def test_update_corp_group_membership(self):
        obj = AutogroupsConfig.objects.create(corp_groups=True)
        obj.states.add(AuthUtils.get_member_state())
        char = EveCharacter.objects.create(
            character_id='1234',
            character_name='test character',
            corporation_id='2345',
            corporation_name='test corp',
            corporation_ticker='tickr',
            alliance_id='3456',
            alliance_name='alliance name',
        )
        self.member.profile.main_character = char
        self.member.profile.save()

        pre_groups = self.member.groups.all()

        # Act
        obj.update_corp_group_membership(self.member)

        group = obj.get_corp_group(self.corp)
        group_qs = Group.objects.filter(pk=group.pk)

        self.assertIn(group, self.member.groups.all())
        self.assertQuerysetEqual(self.member.groups.all(), map(repr, pre_groups | group_qs), ordered=False)

    def test_update_corp_group_membership_no_state(self):
        obj = AutogroupsConfig.objects.create(corp_groups=True)
        char = EveCharacter.objects.create(
            character_id='1234',
            character_name='test character',
            corporation_id='2345',
            corporation_name='test corp',
            corporation_ticker='tickr',
            alliance_id='3456',
            alliance_name='alliance name',
        )
        self.member.profile.main_character = char
        self.member.profile.save()

        pre_groups = list(self.member.groups.all())

        # Act
        obj.update_corp_group_membership(self.member)

        group = obj.get_corp_group(self.corp)

        post_groups = list(self.member.groups.all())

        self.assertNotIn(group, post_groups)
        self.assertListEqual(pre_groups, post_groups)

    def test_update_corp_group_membership_no_main_character(self):
        obj = AutogroupsConfig.objects.create()
        obj.states.add(AuthUtils.get_member_state())

        # Act
        obj.update_corp_group_membership(self.member)

        group = obj.get_corp_group(self.corp)

        self.assertNotIn(group, self.member.groups.all())

    def test_update_corp_group_membership_no_corp_model(self):
        obj = AutogroupsConfig.objects.create()
        obj.states.add(AuthUtils.get_member_state())
        char = EveCharacter.objects.create(
            character_id='1234',
            character_name='test character',
            corporation_id='2348',
            corporation_name='test corp',
            corporation_ticker='tickr',
            alliance_id='3456',
            alliance_name='alliance name',
        )
        self.member.profile.main_character = char
        self.member.profile.save()

        # Act
        obj.update_corp_group_membership(self.member)

        group = obj.get_corp_group(self.corp)

        self.assertNotIn(group, self.member.groups.all())

    def test_remove_user_from_alliance_groups(self):
        obj = AutogroupsConfig.objects.create()
        result = obj.get_alliance_group(self.alliance)

        result.user_set.add(self.member)

        self.assertIn(result, self.member.groups.all())

        # Act
        obj.remove_user_from_alliance_groups(self.member)

        self.assertNotIn(result, self.member.groups.all())

    def test_remove_user_from_corp_groups(self):
        obj = AutogroupsConfig.objects.create()
        result = obj.get_corp_group(self.corp)

        result.user_set.add(self.member)

        self.assertIn(result, self.member.groups.all())

        # Act
        obj.remove_user_from_corp_groups(self.member)

        self.assertNotIn(result, self.member.groups.all())

    def test_get_alliance_group(self):
        obj = AutogroupsConfig.objects.create()
        result = obj.get_alliance_group(self.alliance)

        group = Group.objects.get(name='Alliance alliance name')

        self.assertEqual(result, group)
        self.assertEqual(obj.get_alliance_group_name(self.alliance), result.name)
        self.assertTrue(obj.alliance_managed_groups.filter(pk=result.pk).exists())

    def test_get_corp_group(self):
        obj = AutogroupsConfig.objects.create()
        result = obj.get_corp_group(self.corp)

        group = Group.objects.get(name='Corp corp name')

        self.assertEqual(result, group)
        self.assertEqual(obj.get_corp_group_name(self.corp), group.name)
        self.assertTrue(obj.corp_managed_groups.filter(pk=group.pk).exists())

    def test_create_alliance_group(self):
        obj = AutogroupsConfig.objects.create()
        result = obj.create_alliance_group(self.alliance)

        group = Group.objects.get(name='Alliance alliance name')

        self.assertEqual(result, group)
        self.assertEqual(obj.get_alliance_group_name(self.alliance), group.name)
        self.assertTrue(obj.alliance_managed_groups.filter(pk=group.pk).exists())

    def test_create_corp_group(self):
        obj = AutogroupsConfig.objects.create()
        result = obj.create_corp_group(self.corp)

        group = Group.objects.get(name='Corp corp name')

        self.assertEqual(result, group)
        self.assertEqual(obj.get_corp_group_name(self.corp), group.name)
        self.assertTrue(obj.corp_managed_groups.filter(pk=group.pk).exists())

    def test_delete_alliance_managed_groups(self):
        obj = AutogroupsConfig.objects.create()
        obj.create_alliance_group(self.alliance)

        self.assertTrue(obj.alliance_managed_groups.all().exists())

        obj.delete_alliance_managed_groups()

        self.assertFalse(obj.alliance_managed_groups.all().exists())

    def test_delete_corp_managed_groups(self):
        obj = AutogroupsConfig.objects.create()
        obj.create_corp_group(self.corp)

        self.assertTrue(obj.corp_managed_groups.all().exists())

        obj.delete_corp_managed_groups()

        self.assertFalse(obj.corp_managed_groups.all().exists())

    def test_get_alliance_group_name(self):
        obj = AutogroupsConfig()
        obj.replace_spaces = True
        obj.replace_spaces_with = '_'

        result = obj.get_alliance_group_name(self.alliance)

        self.assertEqual(result, 'Alliance_alliance_name')

    def test_get_alliance_group_name_ticker(self):
        obj = AutogroupsConfig()
        obj.replace_spaces = True
        obj.replace_spaces_with = '_'
        obj.alliance_name_source = obj.OPT_TICKER

        result = obj.get_alliance_group_name(self.alliance)

        self.assertEqual(result, 'Alliance_TIKR')

    def test_get_corp_group_name(self):
        obj = AutogroupsConfig()
        obj.replace_spaces = True
        obj.replace_spaces_with = '_'

        result = obj.get_corp_group_name(self.corp)

        self.assertEqual(result, 'Corp_corp_name')

    def test_get_corp_group_name_ticker(self):
        obj = AutogroupsConfig()
        obj.replace_spaces = True
        obj.replace_spaces_with = '_'
        obj.corp_name_source = obj.OPT_TICKER

        result = obj.get_corp_group_name(self.corp)

        self.assertEqual(result, 'Corp_TIKK')

    def test__replace_spaces(self):
        obj = AutogroupsConfig()
        obj.replace_spaces = True
        obj.replace_spaces_with = '*'
        name = ' test name '

        result = obj._replace_spaces(name)

        self.assertEqual(result, 'test*name')



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-13 04:42
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('optimer', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='optimer',
            name='details',
        ),
        migrations.RemoveField(
            model_name='optimer',
            name='location',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-04 07:17
from __future__ import unicode_literals

from django.db import migrations
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth.management import create_permissions

import logging

logger = logging.getLogger(__name__)


class Migration(migrations.Migration):

    dependencies = [
        ('groupmanagement', '0005_authgroup_public'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='authgroup',
            options={'permissions': (('request_groups', 'Can request non-public groups'),)},
        ),
    ]



from unittest import TestCase

from . import ROLE_ALPHA, ROLE_BRAVO, ROLE_CHARLIE, ROLE_MIKE, ALL_ROLES, create_role
from .. import DiscordRoles


MODULE_PATH = 'allianceauth.services.modules.discord.discord_client.client'


class TestDiscordRoles(TestCase):

    def setUp(self):
        self.all_roles = DiscordRoles(ALL_ROLES)

    def test_can_create_simple(self):
        roles_raw = [ROLE_ALPHA]
        roles = DiscordRoles(roles_raw)
        self.assertListEqual(list(roles), roles_raw)

    def test_can_create_empty(self):
        roles_raw = []
        roles = DiscordRoles(roles_raw)
        self.assertListEqual(list(roles), [])

    def test_raises_exception_if_roles_raw_of_wrong_type(self):
        with self.assertRaises(TypeError):
            DiscordRoles({'id': 1})
    
    def test_raises_exception_if_list_contains_non_dict(self):
        roles_raw = [ROLE_ALPHA, 'not_valid']
        with self.assertRaises(TypeError):
            DiscordRoles(roles_raw)

    def test_raises_exception_if_invalid_role_1(self):
        roles_raw = [{'name': 'alpha', 'managed': False}]
        with self.assertRaises(ValueError):
            DiscordRoles(roles_raw)

    def test_raises_exception_if_invalid_role_2(self):
        roles_raw = [{'id': 1, 'managed': False}]
        with self.assertRaises(ValueError):
            DiscordRoles(roles_raw)

    def test_raises_exception_if_invalid_role_3(self):
        roles_raw = [{'id': 1, 'name': 'alpha'}]
        with self.assertRaises(ValueError):
            DiscordRoles(roles_raw)
    
    def test_roles_are_equal(self):
        roles_a = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_b = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        self.assertEqual(roles_a, roles_b)

    def test_roles_are_not_equal(self):
        roles_a = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_b = DiscordRoles([ROLE_ALPHA])
        self.assertNotEqual(roles_a, roles_b)

    def test_different_objects_are_not_equal(self):
        roles_a = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])        
        self.assertFalse(roles_a == "invalid")

    def test_len(self):        
        self.assertEqual(len(self.all_roles), 4)

    def test_contains(self):
        self.assertTrue(1 in self.all_roles)
        self.assertFalse(99 in self.all_roles)

    def test_sanitize_role_name(self):        
        role_name_input = 'x' * 110
        role_name_expected = 'x' * 100                
        result = DiscordRoles.sanitize_role_name(role_name_input)
        self.assertEqual(result, role_name_expected)

    def test_objects_are_hashable(self):
        roles_a = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_b = DiscordRoles([ROLE_BRAVO, ROLE_ALPHA])
        roles_c = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO, ROLE_MIKE])
        self.assertIsNotNone(hash(roles_a))
        self.assertEqual(hash(roles_a), hash(roles_b))
        self.assertNotEqual(hash(roles_a), hash(roles_c))

    def test_create_from_matched_roles(self):
        matched_roles = [
            (ROLE_ALPHA, True),
            (ROLE_BRAVO, False)
        ]
        roles = DiscordRoles.create_from_matched_roles(matched_roles)
        self.assertSetEqual(roles.ids(), {1, 2})


class TestIds(TestCase):

    def setUp(self):
        self.all_roles = DiscordRoles(ALL_ROLES)
        
    def test_return_role_ids_default(self):        
        result = self.all_roles.ids()
        expected = {1, 2, 3, 13}
        self.assertSetEqual(result, expected)

    def test_return_role_ids_empty(self):
        roles = DiscordRoles([])
        self.assertSetEqual(roles.ids(), set())


class TestSubset(TestCase):

    def setUp(self):
        self.all_roles = DiscordRoles(ALL_ROLES)

    def test_ids_only(self):
        role_ids = {1, 3}
        roles_subset = self.all_roles.subset(role_ids)
        expected = {1, 3}
        self.assertSetEqual(roles_subset.ids(), expected)

    def test_ids_as_string_work_too(self):
        role_ids = {'1', '3'}
        roles_subset = self.all_roles.subset(role_ids)
        expected = {1, 3}
        self.assertSetEqual(roles_subset.ids(), expected)

    def test_managed_only(self):
        roles = self.all_roles.subset(managed_only=True)
        expected = {13}
        self.assertSetEqual(roles.ids(), expected)

    def test_ids_and_managed_only(self):
        role_ids = {1, 3, 13}
        roles_subset = self.all_roles.subset(role_ids, managed_only=True)
        expected = {13}
        self.assertSetEqual(roles_subset.ids(), expected)

    def test_ids_are_empty(self):
        roles = self.all_roles.subset([])
        expected = set()
        self.assertSetEqual(roles.ids(), expected)

    def test_no_parameters(self):
        roles = self.all_roles.subset()
        expected = {1, 2, 3, 13}
        self.assertSetEqual(roles.ids(), expected)


class TestHasRoles(TestCase):

    def setUp(self):
        self.all_roles = DiscordRoles(ALL_ROLES)

    def test_true_if_all_roles_exit(self):
        self.assertTrue(self.all_roles.has_roles([1, 2]))

    def test_true_if_all_roles_exit_str(self):
        self.assertTrue(self.all_roles.has_roles(['1', '2']))
    
    def test_false_if_role_does_not_exit(self):
        self.assertFalse(self.all_roles.has_roles([99]))

    def test_false_if_one_role_does_not_exit(self):
        self.assertFalse(self.all_roles.has_roles([1, 99]))

    def test_true_for_empty_roles(self):
        self.assertTrue(self.all_roles.has_roles([]))
        

class TestGetMatchingRolesByName(TestCase):

    def setUp(self):
        self.all_roles = DiscordRoles(ALL_ROLES)

    def test_return_role_if_matches(self):
        role_name = 'alpha'
        expected = ROLE_ALPHA
        result = self.all_roles.role_by_name(role_name)
        self.assertEqual(result, expected)

    def test_return_role_if_matches_and_limit_max_length(self):
        role_name = 'x' * 120
        expected = create_role(77, 'x' * 100)
        roles = DiscordRoles([expected])
        result = roles.role_by_name(role_name)
        self.assertEqual(result, expected)

    def test_return_empty_if_not_matches(self):
        role_name = 'lima'
        expected = {}
        result = self.all_roles.role_by_name(role_name)
        self.assertEqual(result, expected)


class TestUnion(TestCase):

    def test_distinct_sets(self):
        roles_1 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_2 = DiscordRoles([ROLE_CHARLIE, ROLE_MIKE])
        roles_3 = roles_1.union(roles_2)
        expected = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO, ROLE_CHARLIE, ROLE_MIKE])
        self.assertEqual(roles_3, expected)

    def test_overlapping_sets(self):
        roles_1 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_2 = DiscordRoles([ROLE_BRAVO, ROLE_MIKE])
        roles_3 = roles_1.union(roles_2)
        expected = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO, ROLE_MIKE])
        self.assertEqual(roles_3, expected)

    def test_identical_sets(self):
        roles_1 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_2 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_3 = roles_1.union(roles_2)
        expected = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        self.assertEqual(roles_3, expected)


class TestDifference(TestCase):

    def test_distinct_sets(self):
        roles_1 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_2 = DiscordRoles([ROLE_CHARLIE, ROLE_MIKE])
        roles_3 = roles_1.difference(roles_2)
        expected = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        self.assertEqual(roles_3, expected)

    def test_overlapping_sets(self):
        roles_1 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_2 = DiscordRoles([ROLE_BRAVO, ROLE_MIKE])
        roles_3 = roles_1.difference(roles_2)
        expected = DiscordRoles([ROLE_ALPHA])
        self.assertEqual(roles_3, expected)

    def test_identical_sets(self):
        roles_1 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_2 = DiscordRoles([ROLE_ALPHA, ROLE_BRAVO])
        roles_3 = roles_1.difference(roles_2)
        expected = DiscordRoles([])
        self.assertEqual(roles_3, expected)



from math import ceil
from unittest.mock import patch

from requests import RequestException
import requests_mock
from packaging.version import Version as Pep440Version

from django.test import TestCase

from allianceauth.templatetags.admin_status import (
    status_overview,
    _fetch_list_from_gitlab,
    _current_notifications, 
    _current_version_summary, 
    _fetch_notification_issues_from_gitlab,
    _fetch_tags_from_gitlab,
    _latests_versions
)

MODULE_PATH = 'allianceauth.templatetags'


def create_tags_list(tag_names: list):
    return [{'name': str(tag_name)} for tag_name in tag_names]


GITHUB_TAGS = create_tags_list(['v2.4.6a1', 'v2.4.5', 'v2.4.0', 'v2.0.0', 'v1.1.1'])
GITHUB_NOTIFICATION_ISSUES = [
    {
        'id': 1,
        'title': 'first issue'
    },
    {
        'id': 2,
        'title': 'second issue'
    },
    {
        'id': 3,
        'title': 'third issue'
    },
    {
        'id': 4,
        'title': 'forth issue'
    },
    {
        'id': 5,
        'title': 'fifth issue'
    },
    {
        'id': 6,
        'title': 'sixth issue'
    },
]
TEST_VERSION = '2.6.5'


class TestStatusOverviewTag(TestCase):

    @patch(MODULE_PATH + '.admin_status.__version__', TEST_VERSION)
    @patch(MODULE_PATH + '.admin_status._fetch_celery_queue_length')
    @patch(MODULE_PATH + '.admin_status._current_version_summary')
    @patch(MODULE_PATH + '.admin_status._current_notifications')    
    def test_status_overview(
        self, 
        mock_current_notifications, 
        mock_current_version_info,
        mock_fetch_celery_queue_length
    ):
        notifications = {
            'notifications': GITHUB_NOTIFICATION_ISSUES[:5]
        }
        mock_current_notifications.return_value = notifications
        version_info = {
            'latest_major': True,
            'latest_minor': True,
            'latest_patch': True,
            'latest_beta': False,
            'current_version': TEST_VERSION,
            'latest_major_version': '2.4.5',
            'latest_minor_version': '2.4.0',
            'latest_patch_version': '2.4.5',
            'latest_beta_version': '2.4.4a1',
        }
        mock_current_version_info.return_value = version_info
        mock_fetch_celery_queue_length.return_value = 3
                
        result = status_overview()
        expected = {
            'notifications': GITHUB_NOTIFICATION_ISSUES[:5],
            'latest_major': True,
            'latest_minor': True,
            'latest_patch': True,
            'latest_beta': False,
            'current_version': TEST_VERSION,
            'latest_major_version': '2.4.5',
            'latest_minor_version': '2.4.0',
            'latest_patch_version': '2.4.5',
            'latest_beta_version': '2.4.4a1',
            'task_queue_length': 3,
        }
        self.assertEqual(result, expected)


class TestNotifications(TestCase):

    @requests_mock.mock()
    def test_fetch_notification_issues_from_gitlab(self, requests_mocker):
        url = (
            'https://gitlab.com/api/v4/projects/allianceauth%2Fallianceauth/issues'
            '?labels=announcement'
        )        
        requests_mocker.get(url, json=GITHUB_NOTIFICATION_ISSUES)
        result = _fetch_notification_issues_from_gitlab()
        self.assertEqual(result, GITHUB_NOTIFICATION_ISSUES)

    @patch(MODULE_PATH + '.admin_status.cache')
    def test_current_notifications_normal(self, mock_cache):
        mock_cache.get_or_set.return_value = GITHUB_NOTIFICATION_ISSUES

        result = _current_notifications()
        self.assertEqual(result['notifications'], GITHUB_NOTIFICATION_ISSUES[:5])

    @patch(MODULE_PATH + '.admin_status.cache')
    def test_current_notifications_failed(self, mock_cache):
        mock_cache.get_or_set.side_effect = RequestException

        result = _current_notifications()
        self.assertEqual(result['notifications'], list())

    @patch(MODULE_PATH + '.admin_status.cache')
    def test_current_notifications_is_none(self, mock_cache):
        mock_cache.get_or_set.return_value = None

        result = _current_notifications()
        self.assertEqual(result['notifications'], list())


class TestCeleryQueueLength(TestCase):

    def test_get_celery_queue_length(self):
        pass


class TestVersionTags(TestCase):

    @patch(MODULE_PATH + '.admin_status.__version__', TEST_VERSION)
    @patch(MODULE_PATH + '.admin_status.cache')
    def test_current_version_info_normal(self, mock_cache):
        mock_cache.get_or_set.return_value = GITHUB_TAGS

        result = _current_version_summary()
        self.assertTrue(result['latest_major'])
        self.assertTrue(result['latest_minor'])
        self.assertTrue(result['latest_patch'])
        self.assertEqual(result['latest_major_version'], '2.0.0')
        self.assertEqual(result['latest_minor_version'], '2.4.0')
        self.assertEqual(result['latest_patch_version'], '2.4.5')
        self.assertEqual(result['latest_beta_version'], '2.4.6a1')

    @patch(MODULE_PATH + '.admin_status.__version__', TEST_VERSION)
    @patch(MODULE_PATH + '.admin_status.cache')
    def test_current_version_info_failed(self, mock_cache):
        mock_cache.get_or_set.side_effect = RequestException
        
        expected = {}
        result = _current_version_summary()
        self.assertEqual(result, expected)

    @requests_mock.mock()
    def test_fetch_tags_from_gitlab(self, requests_mocker):
        url = (
            'https://gitlab.com/api/v4/projects/allianceauth%2Fallianceauth'
            '/repository/tags'
        )        
        requests_mocker.get(url, json=GITHUB_TAGS)
        result = _fetch_tags_from_gitlab()
        self.assertEqual(result, GITHUB_TAGS)

    @patch(MODULE_PATH + '.admin_status.__version__', TEST_VERSION)
    @patch(MODULE_PATH + '.admin_status.cache')
    def test_current_version_info_return_no_data(self, mock_cache):
        mock_cache.get_or_set.return_value = None
        
        expected = {}
        result = _current_version_summary()
        self.assertEqual(result, expected)


class TestLatestsVersion(TestCase):

    def test_all_version_types_defined(self):
        
        tags = create_tags_list(
            ['2.1.1', '2.1.0', '2.0.0', '2.1.1a1', '1.1.1', '1.1.0', '1.0.0']
        )
        major, minor, patch, beta = _latests_versions(tags)
        self.assertEqual(major, Pep440Version('2.0.0'))
        self.assertEqual(minor, Pep440Version('2.1.0'))
        self.assertEqual(patch, Pep440Version('2.1.1'))
        self.assertEqual(beta, Pep440Version('2.1.1a1'))

    def test_major_and_minor_not_defined_with_zero(self):
        
        tags = create_tags_list(
            ['2.1.2', '2.1.1', '2.0.1', '2.1.1a1', '1.1.1', '1.1.0', '1.0.0']
        )
        major, minor, patch, beta = _latests_versions(tags)
        self.assertEqual(major, Pep440Version('2.0.1'))
        self.assertEqual(minor, Pep440Version('2.1.1'))
        self.assertEqual(patch, Pep440Version('2.1.2'))
        self.assertEqual(beta, Pep440Version('2.1.1a1'))
        
    def test_can_ignore_invalid_versions(self):
        
        tags = create_tags_list(
            ['2.1.1', '2.1.0', '2.0.0', '2.1.1a1', 'invalid']
        )
        major, minor, patch, beta = _latests_versions(tags)
        self.assertEqual(major, Pep440Version('2.0.0'))
        self.assertEqual(minor, Pep440Version('2.1.0'))
        self.assertEqual(patch, Pep440Version('2.1.1'))
        self.assertEqual(beta, Pep440Version('2.1.1a1'))


class TestFetchListFromGitlab(TestCase):
   
    page_size = 2
    
    def setUp(self):
        self.url = (
            'https://gitlab.com/api/v4/projects/allianceauth%2Fallianceauth'
            '/repository/tags'
        )

    @classmethod
    def my_callback(cls, request, context):
        page = int(request.qs['page'][0])
        start = (page - 1) * cls.page_size
        end = start + cls.page_size
        return GITHUB_TAGS[start:end]        
    
    @requests_mock.mock()
    def test_can_fetch_one_page_with_header(self, requests_mocker):
        headers = {
            'x-total-pages': '1'
        }
        requests_mocker.get(self.url, json=GITHUB_TAGS, headers=headers)
        result = _fetch_list_from_gitlab(self.url)
        self.assertEqual(result, GITHUB_TAGS)
        self.assertEqual(requests_mocker.call_count, 1)

    @requests_mock.mock()
    def test_can_fetch_one_page_wo_header(self, requests_mocker):        
        requests_mocker.get(self.url, json=GITHUB_TAGS)
        result = _fetch_list_from_gitlab(self.url)
        self.assertEqual(result, GITHUB_TAGS)
        self.assertEqual(requests_mocker.call_count, 1)

    @requests_mock.mock()
    def test_can_fetch_one_page_and_ignore_invalid_header(self, requests_mocker):
        headers = {
            'x-total-pages': 'invalid'
        }
        requests_mocker.get(self.url, json=GITHUB_TAGS, headers=headers)
        result = _fetch_list_from_gitlab(self.url)
        self.assertEqual(result, GITHUB_TAGS)
        self.assertEqual(requests_mocker.call_count, 1)

    @requests_mock.mock()
    def test_can_fetch_multiple_pages(self, requests_mocker):     
        total_pages = ceil(len(GITHUB_TAGS) / self.page_size)
        headers = {
            'x-total-pages': str(total_pages)
        }
        requests_mocker.get(self.url, json=self.my_callback, headers=headers)
        result = _fetch_list_from_gitlab(self.url)
        self.assertEqual(result, GITHUB_TAGS)
        self.assertEqual(requests_mocker.call_count, total_pages)

    @requests_mock.mock()
    def test_can_fetch_given_number_of_pages_only(self, requests_mocker):     
        total_pages = ceil(len(GITHUB_TAGS) / self.page_size)
        headers = {
            'x-total-pages': str(total_pages)
        }
        requests_mocker.get(self.url, json=self.my_callback, headers=headers)
        max_pages = 2
        result = _fetch_list_from_gitlab(self.url, max_pages=max_pages)
        self.assertEqual(result, GITHUB_TAGS[:4])
        self.assertEqual(requests_mocker.call_count, max_pages)



# Generated by Django 2.0.4 on 2018-04-17 20:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0009_on_delete'),
    ]

    operations = [
        migrations.AddField(
            model_name='evecharacter',
            name='alliance_ticker',
            field=models.CharField(blank=True, default='', max_length=5, null=True),
        ),
        migrations.AlterField(
            model_name='evecharacter',
            name='corporation_ticker',
            field=models.CharField(max_length=5),
        ),
    ]



from copy import copy


class DiscordRoles:
    """Container class that helps dealing with Discord roles.
    
    Objects of this class are immutable and work in many ways like sets.
    
    Ideally objects are initialized from raw API responses, 
    e.g. from DiscordClient.guild.roles()
    """
    _ROLE_NAME_MAX_CHARS = 100

    def __init__(self, roles_lst: list) -> None:
        """roles_lst must be a list of dict, each defining a role"""
        if not isinstance(roles_lst, (list, set, tuple)):
            raise TypeError('roles_lst must be of type list, set or tuple')
        self._roles = dict()
        self._roles_by_name = dict()
        for role in list(roles_lst):
            self._assert_valid_role(role)
            self._roles[int(role['id'])] = role
            self._roles_by_name[self.sanitize_role_name(role['name'])] = role
        
    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.ids() == other.ids()
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self._roles.keys())))
    
    def __iter__(self):
        for role in self._roles.values():
            yield role

    def __contains__(self, item) -> bool:        
        return int(item) in self._roles

    def __len__(self):        
        return len(self._roles.keys())
    
    def has_roles(self, role_ids: set) -> bool:
        """returns true if this objects contains all roles defined by given role_ids
        incl. managed roles
        """
        role_ids = {int(id) for id in role_ids}
        all_role_ids = self._roles.keys()        
        return role_ids.issubset(all_role_ids)
    
    def ids(self) -> set:
        """return a set of all role IDs"""
        return set(self._roles.keys())

    def subset(self, role_ids: set = None, managed_only: bool = False) -> object:
        """returns a new object containing the subset of roles as defined 
        by given role IDs and/or including managed roles only
        """
        if role_ids is not None:
            role_ids = {int(id) for id in role_ids}

        if role_ids is not None and not managed_only:
            return type(self)([
                role for role_id, role in self._roles.items() if role_id in role_ids
            ])

        elif role_ids is None and managed_only:
            return type(self)([
                role for _, role in self._roles.items() if role['managed']
            ])
        
        elif role_ids is not None and managed_only:
            return type(self)([
                role for role_id, role in self._roles.items() 
                if role_id in role_ids and role['managed']
            ])
        
        else:
            return copy(self)

    def union(self, other: object) -> object:
        """returns a new roles object that is the union of this roles object 
        with other"""
        return type(self)(list(self) + list(other))

    def difference(self, other: object) -> object:
        """returns a new roles object that only contains the roles 
        that exist in the current objects, but not in other
        """
        new_ids = self.ids().difference(other.ids())
        return self.subset(role_ids=new_ids)

    def role_by_name(self, role_name: str) -> dict:
        """returns role if one with matching name is found else an empty dict"""
        role_name = self.sanitize_role_name(role_name)
        if role_name in self._roles_by_name:
            return self._roles_by_name[role_name]
        else:
            return dict()
        
    @classmethod
    def create_from_matched_roles(cls, matched_roles: list) -> None:
        """returns a new object created from the given list of matches roles
        
        matches_roles must be a list of tuples in the form: (role, created)
        """
        raw_roles = [x[0] for x in matched_roles]
        return cls(raw_roles)

    @staticmethod
    def _assert_valid_role(role: dict):
        if not isinstance(role, dict):
            raise TypeError('Roles must be of type dict: %s' % role)
            
        if 'id' not in role or 'name' not in role or 'managed' not in role:
            raise ValueError('This role is not valid: %s' % role)
    
    @classmethod
    def sanitize_role_name(cls, role_name: str) -> str:
        """shortens too long strings if necessary"""
        return str(role_name)[:cls._ROLE_NAME_MAX_CHARS]


def match_or_create_roles_from_names(
    client: object, guild_id: int, role_names: list
) -> DiscordRoles:
    """Shortcut for getting the result of matching role names as DiscordRoles object"""
    return DiscordRoles.create_from_matched_roles(
        client.match_or_create_roles_from_names(
            guild_id=guild_id, role_names=role_names
        )
    )



# this package generates profile URL for eve entities 
# on 3rd party websites like evewho and zKillboard
#
# It contains of modules for views and templatetags for templates

# list of all eve entity categories as defined in ESI
_ESI_CATEGORY_AGENT = "agent"
_ESI_CATEGORY_ALLIANCE = "alliance"
_ESI_CATEGORY_CHARACTER = "character"
_ESI_CATEGORY_CONSTELLATION = "constellation"
_ESI_CATEGORY_CORPORATION = "corporation"
_ESI_CATEGORY_FACTION = "faction"
_ESI_CATEGORY_INVENTORYTYPE = "inventory_type"
_ESI_CATEGORY_REGION = "region"
_ESI_CATEGORY_SOLARSYSTEM = "solar_system"
_ESI_CATEGORY_STATION = "station"
_ESI_CATEGORY_WORMHOLE = "wormhole"



from django.contrib.auth.models import User
from django.db import models


class DiscourseUser(models.Model):
    user = models.OneToOneField(User,
                                primary_key=True,
                                on_delete=models.CASCADE,
                                related_name='discourse')
    enabled = models.BooleanField()

    def __str__(self):
        return self.user.username

    class Meta:
        permissions = (
            ("access_discourse", u"Can access the Discourse service"),
        )




from urllib.parse import quote
from unittest.mock import patch, MagicMock

from django.conf import settings
from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User as BaseUser, Group
from django.test import TestCase, RequestFactory, Client

from allianceauth.authentication.models import (
    CharacterOwnership, State, OwnershipRecord
)
from allianceauth.eveonline.models import (
    EveCharacter, EveCorporationInfo, EveAllianceInfo
)
from allianceauth.services.hooks import ServicesHook
from allianceauth.tests.auth_utils import AuthUtils

from ..admin import (
    BaseUserAdmin,
    CharacterOwnershipAdmin,
    PermissionAdmin,
    StateAdmin,
    MainCorporationsFilter,
    MainAllianceFilter,
    OwnershipRecordAdmin,
    User,
    UserAdmin,     
    user_main_organization,
    user_profile_pic, 
    user_username,
    update_main_character_model,
    make_service_hooks_update_groups_action,
    make_service_hooks_sync_nickname_action
)
from . import get_admin_change_view_url, get_admin_search_url

if 'allianceauth.eveonline.autogroups' in settings.INSTALLED_APPS:
    _has_auto_groups = True
    from allianceauth.eveonline.autogroups.models import AutogroupsConfig
else:
    _has_auto_groups = False

MODULE_PATH = 'allianceauth.authentication.admin'


class MockRequest(object):    
    def __init__(self, user=None):
        self.user = user

class TestCaseWithTestData(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        for MyModel in [
            EveAllianceInfo, EveCorporationInfo, EveCharacter, Group, User
        ]:
            MyModel.objects.all().delete()
        
        # groups
        cls.group_1 = Group.objects.create(
            name='Group 1'
        )
        cls.group_2 = Group.objects.create(
            name='Group 2'
        )

        # user 1 - corp and alliance, normal user
        character_1 = EveCharacter.objects.create(
            character_id=1001,
            character_name='Bruce Wayne',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        character_1a = EveCharacter.objects.create(
            character_id=1002,
            character_name='Batman',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        alliance = EveAllianceInfo.objects.create(
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',            
            executor_corp_id=2001
        )
        EveCorporationInfo.objects.create(
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',            
            member_count=42,
            alliance=alliance
        )        
        cls.user_1 = User.objects.create_user(
            character_1.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=character_1,
            owner_hash='x1' + character_1.character_name,
            user=cls.user_1
        )
        CharacterOwnership.objects.create(
            character=character_1a,
            owner_hash='x1' + character_1a.character_name,
            user=cls.user_1
        )
        cls.user_1.profile.main_character = character_1
        cls.user_1.profile.save()
        cls.user_1.groups.add(cls.group_1)        

        # user 2 - corp only, staff
        character_2 = EveCharacter.objects.create(
            character_id=1003,
            character_name='Clark Kent',
            corporation_id=2002,
            corporation_name='Daily Planet',
            corporation_ticker='DP',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2002,
            corporation_name='Daily Plane',
            corporation_ticker='DP',            
            member_count=99,
            alliance=None
        )
        cls.user_2 = User.objects.create_user(
            character_2.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=character_2,
            owner_hash='x1' + character_2.character_name,
            user=cls.user_2
        )
        cls.user_2.profile.main_character = character_2
        cls.user_2.profile.save()
        cls.user_2.groups.add(cls.group_2)
        cls.user_2.is_staff = True
        cls.user_2.save()
        
        # user 3 - no main, no group, superuser
        character_3 = EveCharacter.objects.create(
            character_id=1101,
            character_name='Lex Luthor',
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',            
            member_count=666,
            alliance=None
        )
        EveAllianceInfo.objects.create(
            alliance_id=3101,
            alliance_name='Lex World Domination',
            alliance_ticker='LWD',
            executor_corp_id=2101
        )
        cls.user_3 = User.objects.create_user(
            character_3.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=character_3,
            owner_hash='x1' + character_3.character_name,
            user=cls.user_3
        )
        cls.user_3.is_superuser = True
        cls.user_3.save()


def make_generic_search_request(ModelClass: type, search_term: str):
    User.objects.create_superuser(
        username='superuser', password='secret', email='admin@example.com'
    )
    c = Client()
    c.login(username='superuser', password='secret')    
    return c.get(
        '%s?q=%s' % (get_admin_search_url(ModelClass), quote(search_term))
    )    


class TestCharacterOwnershipAdmin(TestCaseWithTestData):
    
    def setUp(self):
        self.modeladmin = CharacterOwnershipAdmin(
            model=User, admin_site=AdminSite()
        )

    def test_change_view_loads_normally(self):        
        User.objects.create_superuser(
            username='superuser', password='secret', email='admin@example.com'
        )
        c = Client()
        c.login(username='superuser', password='secret')                
        ownership = self.user_1.character_ownerships.first()
        response = c.get(get_admin_change_view_url(ownership))
        self.assertEqual(response.status_code, 200)

    def test_search_works(self):
        obj = CharacterOwnership.objects\
            .filter(user=self.user_1)\
            .first()
        response = make_generic_search_request(type(obj), obj.user.username)
        expected = 200
        self.assertEqual(response.status_code, expected)


class TestOwnershipRecordAdmin(TestCaseWithTestData):
    
    def setUp(self):
        self.modeladmin = OwnershipRecordAdmin(
            model=User, admin_site=AdminSite()
        )

    def test_change_view_loads_normally(self):        
        User.objects.create_superuser(
            username='superuser', password='secret', email='admin@example.com'
        )
        c = Client()
        c.login(username='superuser', password='secret')                
        ownership_record = OwnershipRecord.objects\
            .filter(user=self.user_1)\
            .first()
        response = c.get(get_admin_change_view_url(ownership_record))
        self.assertEqual(response.status_code, 200)

    def test_search_works(self):
        obj = OwnershipRecord.objects.first()
        response = make_generic_search_request(type(obj), obj.user.username)
        expected = 200
        self.assertEqual(response.status_code, expected)


class TestStateAdmin(TestCaseWithTestData):
        
    def setUp(self):
        self.modeladmin = StateAdmin(
            model=User, admin_site=AdminSite()
        )

    def test_change_view_loads_normally(self):        
        User.objects.create_superuser(
            username='superuser', password='secret', email='admin@example.com'
        )
        c = Client()
        c.login(username='superuser', password='secret')                
        
        guest_state = AuthUtils.get_guest_state()
        response = c.get(get_admin_change_view_url(guest_state))
        self.assertEqual(response.status_code, 200)
        
        member_state = AuthUtils.get_member_state()
        response = c.get(get_admin_change_view_url(member_state))
        self.assertEqual(response.status_code, 200)

    def test_search_works(self):
        obj = State.objects.first()
        response = make_generic_search_request(type(obj), obj.name)
        expected = 200
        self.assertEqual(response.status_code, expected)

class TestUserAdmin(TestCaseWithTestData):

    def setUp(self):
        self.factory = RequestFactory()
        self.modeladmin = UserAdmin(
            model=User, admin_site=AdminSite()
        )
        self.character_1 = self.user_1.character_ownerships.first().character

    def _create_autogroups(self):
        """create autogroups for corps and alliances"""
        if _has_auto_groups:
            autogroups_config = AutogroupsConfig(
                corp_groups = True,
                alliance_groups = True
            )
            autogroups_config.save()
            for state in State.objects.all():
                autogroups_config.states.add(state)        
            autogroups_config.update_corp_group_membership(self.user_1)        
    
    # column rendering

    def test_user_profile_pic_u1(self):
        expected = ('<img src="https://images.evetech.net/characters/1001/'
            'portrait?size=32" class="img-circle">')        
        self.assertEqual(user_profile_pic(self.user_1), expected)

    def test_user_profile_pic_u3(self):
        self.assertIsNone(user_profile_pic(self.user_3))

    def test_user_username_u1(self):
        expected = (
            '<strong><a href="/admin/authentication/user/{}/change/">'
            'Bruce_Wayne</a></strong><br>Bruce Wayne'.format(self.user_1.pk)
        )
        self.assertEqual(user_username(self.user_1), expected)

    def test_user_username_u3(self):
        expected = (
            '<strong><a href="/admin/authentication/user/{}/change/">'
            'Lex_Luthor</a></strong>'.format(self.user_3.pk)
        )
        self.assertEqual(user_username(self.user_3), expected)

    def test_user_main_organization_u1(self):
        expected = 'Wayne Technologies<br>Wayne Enterprises'
        self.assertEqual(user_main_organization(self.user_1), expected)

    def test_user_main_organization_u2(self):
        expected = 'Daily Planet'
        self.assertEqual(user_main_organization(self.user_2), expected)

    def test_user_main_organization_u3(self):
        expected = None
        self.assertEqual(user_main_organization(self.user_3), expected)

    def test_characters_u1(self):
        expected = 'Batman, Bruce Wayne'
        result = self.modeladmin._characters(self.user_1)
        self.assertEqual(result, expected)

    def test_characters_u2(self):
        expected = 'Clark Kent'
        result = self.modeladmin._characters(self.user_2)
        self.assertEqual(result, expected)

    def test_characters_u3(self):
        expected = 'Lex Luthor'
        result = self.modeladmin._characters(self.user_3)
        self.assertEqual(result, expected)
    
    def test_groups_u1(self):
        self._create_autogroups()
        expected = 'Group 1'
        result = self.modeladmin._groups(self.user_1)
        self.assertEqual(result, expected)

    def test_groups_u2(self):
        self._create_autogroups()
        expected = 'Group 2'
        result = self.modeladmin._groups(self.user_2)
        self.assertEqual(result, expected)
    
    def test_groups_u3(self):        
        self._create_autogroups()
        result = self.modeladmin._groups(self.user_3)
        self.assertIsNone(result)

    @patch(MODULE_PATH + '._has_auto_groups', False)
    def test_groups_u1_no_autogroups(self):        
        expected = 'Group 1'
        result = self.modeladmin._groups(self.user_1)
        self.assertEqual(result, expected)

    @patch(MODULE_PATH + '._has_auto_groups', False)
    def test_groups_u2_no_autogroups(self):        
        expected = 'Group 2'
        result = self.modeladmin._groups(self.user_2)
        self.assertEqual(result, expected)
    
    @patch(MODULE_PATH + '._has_auto_groups', False)
    def test_groups_u3_no_autogroups(self):                
        result = self.modeladmin._groups(self.user_3)
        self.assertIsNone(result)

    def test_state(self):
        expected = 'Guest'
        result = self.modeladmin._state(self.user_1)
        self.assertEqual(result, expected)

    def test_role_u1(self):
        expected = 'User'
        result = self.modeladmin._role(self.user_1)
        self.assertEqual(result, expected)

    def test_role_u2(self):
        expected = 'Staff'
        result = self.modeladmin._role(self.user_2)
        self.assertEqual(result, expected)

    def test_role_u3(self):
        expected = 'Superuser'
        result = self.modeladmin._role(self.user_3)
        self.assertEqual(result, expected)

    def test_list_2_html_w_tooltips_no_cutoff(self):
        items = ['one', 'two', 'three']
        expected = 'one, two, three'
        result = self.modeladmin._list_2_html_w_tooltips(items, 5)
        self.assertEqual(expected, result)

    def test_list_2_html_w_tooltips_w_cutoff(self):
        items = ['one', 'two', 'three']
        expected = ('<span data-tooltip="one, two, three" '
            'class="tooltip">one, two, (...)</span>')
        result = self.modeladmin._list_2_html_w_tooltips(items, 2)
        self.assertEqual(expected, result)

    def test_list_2_html_w_tooltips_empty_list(self):
        items = []
        expected = None
        result = self.modeladmin._list_2_html_w_tooltips(items, 5)
        self.assertEqual(expected, result)
    
    # actions

    @patch(MODULE_PATH + '.UserAdmin.message_user', auto_spec=True)
    @patch(MODULE_PATH + '.update_character')
    def test_action_update_main_character_model(
        self, mock_task, mock_message_user
    ):
        users_qs = User.objects.filter(pk__in=[self.user_1.pk, self.user_2.pk])
        update_main_character_model(
            self.modeladmin, MockRequest(self.user_1), users_qs
        )
        self.assertEqual(mock_task.delay.call_count, 2)
        self.assertTrue(mock_message_user.called)
    
    # filters

    def test_filter_real_groups_with_autogroups(self):
        
        class UserAdminTest(BaseUserAdmin): 
            list_filter = (UserAdmin.RealGroupsFilter,)
                
        self._create_autogroups()        
        my_modeladmin = UserAdminTest(User, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [
            (self.group_1.pk, self.group_1.name), 
            (self.group_2.pk, self.group_2.name),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get('/', {'group_id__exact': self.group_1.pk})
        request.user = self.user_1                
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = User.objects.filter(groups__in=[self.group_1])
        self.assertSetEqual(set(queryset), set(expected))

    @patch(MODULE_PATH + '._has_auto_groups', False)
    def test_filter_real_groups_no_autogroups(self):
        
        class UserAdminTest(BaseUserAdmin): 
            list_filter = (UserAdmin.RealGroupsFilter,)
                        
        my_modeladmin = UserAdminTest(User, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [
            (self.group_1.pk, self.group_1.name), 
            (self.group_2.pk, self.group_2.name),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get('/', {'group_id__exact': self.group_1.pk})
        request.user = self.user_1                
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = User.objects.filter(groups__in=[self.group_1])
        self.assertSetEqual(set(queryset), set(expected))

    def test_filter_main_corporations(self):
        
        class UserAdminTest(BaseUserAdmin): 
            list_filter = (MainCorporationsFilter,)
                
        my_modeladmin = UserAdminTest(User, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [            
            (2002, 'Daily Planet'),
            (2001, 'Wayne Technologies'),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get(
            '/', 
            {'main_corporation_id__exact': self.character_1.corporation_id}
        )
        request.user = self.user_1                
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = [self.user_1]
        self.assertSetEqual(set(queryset), set(expected))

    def test_filter_main_alliances(self):
        
        class UserAdminTest(BaseUserAdmin): 
            list_filter = (MainAllianceFilter,)
                
        my_modeladmin = UserAdminTest(User, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [            
            (3001, 'Wayne Enterprises'),            
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get(
            '/', 
            {'main_alliance_id__exact': self.character_1.alliance_id}
        )
        request.user = self.user_1                
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = [self.user_1]
        self.assertSetEqual(set(queryset), set(expected))

    def test_change_view_loads_normally(self):
        User.objects.create_superuser(
            username='superuser', password='secret', email='admin@example.com'
        )
        c = Client()
        c.login(username='superuser', password='secret')                
        response = c.get(get_admin_change_view_url(self.user_1))
        self.assertEqual(response.status_code, 200)

    def test_search_works(self):
        obj = User.objects.first()
        response = make_generic_search_request(type(obj), obj.username)
        expected = 200
        self.assertEqual(response.status_code, expected)


class TestMakeServicesHooksActions(TestCaseWithTestData):

    class MyServicesHookTypeA(ServicesHook):

        def __init__(self):
            super().__init__()
            self.name = 'My Service A'            
        
        def update_groups(self, user):
            pass

        def sync_nicknames(self, user):
            pass

    class MyServicesHookTypeB(ServicesHook):

        def __init__(self):
            super().__init__()
            self.name = 'My Service B'
        
        def update_groups(self, user):
            pass

        def update_groups_bulk(self, user):
            pass

        def sync_nicknames(self, user):
            pass

        def sync_nicknames_bulk(self, user):
            pass
         

    def test_service_has_update_groups_only(self):        
        service = self.MyServicesHookTypeA()
        mock_service = MagicMock(spec=service)      
        action = make_service_hooks_update_groups_action(mock_service)
        action(MagicMock(), MagicMock(), [self.user_1])
        self.assertTrue(mock_service.update_groups.called)

    def test_service_has_update_groups_bulk(self):        
        service = self.MyServicesHookTypeB()
        mock_service = MagicMock(spec=service)      
        action = make_service_hooks_update_groups_action(mock_service)
        action(MagicMock(), MagicMock(), [self.user_1])
        self.assertFalse(mock_service.update_groups.called)
        self.assertTrue(mock_service.update_groups_bulk.called)

    def test_service_has_sync_nickname_only(self):        
        service = self.MyServicesHookTypeA()
        mock_service = MagicMock(spec=service)      
        action = make_service_hooks_sync_nickname_action(mock_service)
        action(MagicMock(), MagicMock(), [self.user_1])
        self.assertTrue(mock_service.sync_nickname.called)

    def test_service_has_sync_nicknames_bulk(self):        
        service = self.MyServicesHookTypeB()
        mock_service = MagicMock(spec=service)      
        action = make_service_hooks_sync_nickname_action(mock_service)
        action(MagicMock(), MagicMock(), [self.user_1])
        self.assertFalse(mock_service.sync_nickname.called)
        self.assertTrue(mock_service.sync_nicknames_bulk.called)



from django.conf.urls import include, url
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.conf import settings
from string import Formatter

from allianceauth.hooks import get_hooks

from .models import NameFormatConfig


def get_extension_logger(name):
    """
    Takes the name of a plugin/extension and generates a child logger of the extensions logger
    to be used by the extension to log events to the extensions logger.

    The logging level is determined by the level defined for the parent logger.

    :param: name: the name of the extension doing the logging
    :return: an extensions child logger
    """
    if not isinstance(name, str):
        raise TypeError(f"get_extension_logger takes an argument of type string."
                        f"Instead received argument of type {type(name).__name__}.")

    import logging

    parent_logger = logging.getLogger('extensions')

    logger = logging.getLogger('extensions.' + name)
    logger.name = name
    logger.level = parent_logger.level

    return logger


class ServicesHook:
    """
    Abstract base class for creating a compatible services
    hook. Decorate with @register('services_hook') to have the
    services module registered for callbacks. Must be in
    auth_hook(.py) sub module
    """
    def __init__(self):
        self.name = 'Undefined'
        self.urlpatterns = []
        self.service_ctrl_template = 'services/services_ctrl.html'
        self.access_perm = None

    @property
    def title(self):
        """
        A nicely formatted title of the service, for client facing
        display.
        :return: str
        """
        return self.name.title()

    def delete_user(self, user, notify_user=False):
        """
        Delete the users service account, optionally notify them
        that the service has been disabled
        :param user: Django.contrib.auth.models.User
        :param notify_user: Whether the service should sent a
        notification to the user about the disabling of their
        service account.
        :return: True if the service account has been disabled,
        or False if it doesnt exist.
        """
        pass

    def validate_user(self, user):
        pass

    def sync_nickname(self, user):
        """
        Sync the users nickname
        :param user: Django.contrib.auth.models.User
        :return: None
        """
        pass

    def update_groups(self, user):
        """
        Update the users group membership
        :param user: Django.contrib.auth.models.User
        :return: None
        """
        pass

    def update_all_groups(self):
        """
        Iterate through and update all users groups
        :return: None
        """
        pass

    def service_active_for_user(self, user):
        pass

    def show_service_ctrl(self, user):
        """
        Whether the service control should be displayed to the given user
        who has the given service state. Usually this function wont
        require overloading.
        :param user: django.contrib.auth.models.User
        :return: bool True if the service should be shown
        """
        return self.service_active_for_user(user) or user.is_superuser

    def render_services_ctrl(self, request):
        """
        Render the services control template row
        :param request:
        :return:
        """
        return ''

    def __str__(self):
        return self.name or 'Unknown Service Module'

    class Urls:
        def __init__(self):
            self.auth_activate = ''
            self.auth_set_password = ''
            self.auth_reset_password = ''
            self.auth_deactivate = ''

    @staticmethod
    def get_services():
        for fn in get_hooks('services_hook'):
            yield fn()


class MenuItemHook:
    def __init__(self, text, classes, url_name, order=None, navactive=list([])):
        self.text = text
        self.classes = classes
        self.url_name = url_name
        self.template = 'public/menuitem.html'
        self.order = order if order is not None else 9999
        
        # count is an integer shown next to the menu item as badge when count != None
        # apps need to set the count in their child class, e.g. in render() method
        self.count = None
        
        navactive = navactive or []
        navactive.append(url_name)
        self.navactive = navactive

    def render(self, request):
        return render_to_string(self.template,
                                {'item': self},
                                request=request)


class UrlHook:
    def __init__(self, urls, namespace, base_url):
        self.include_pattern = url(base_url, include(urls, namespace=namespace))


class NameFormatter:
    DEFAULT_FORMAT = getattr(settings, "DEFAULT_SERVICE_NAME_FORMAT", '[{corp_ticker}] {character_name}')

    def __init__(self, service, user):
        """
        :param service: ServicesHook of the service to generate the name for.
        :param user: django.contrib.auth.models.User to format name for
        """
        self.service = service
        self.user = user

    def format_name(self):
        """
        :return: str Generated name
        """
        format_data = self.get_format_data()
        return Formatter().vformat(self.string_formatter, args=[], kwargs=format_data)

    def get_format_data(self):
        main_char = getattr(self.user.profile, 'main_character', None)

        format_data = {
            'character_name': getattr(main_char, 'character_name',
                                      self.user.username if self._default_to_username else None),
            'character_id': getattr(main_char, 'character_id', None),
            'corp_ticker': getattr(main_char, 'corporation_ticker', None),
            'corp_name': getattr(main_char, 'corporation_name', None),
            'corp_id': getattr(main_char, 'corporation_id', None),
            'alliance_name': getattr(main_char, 'alliance_name', None),
            'alliance_id': getattr(main_char, 'alliance_id', None),
            'alliance_ticker': getattr(main_char, 'alliance_ticker', None),
            'username': self.user.username,
        }

        format_data['alliance_or_corp_name'] = format_data['alliance_name'] or format_data['corp_name']
        format_data['alliance_or_corp_ticker'] = format_data['alliance_ticker'] or format_data['corp_ticker']
        return format_data

    @cached_property
    def formatter_config(self):
        format_config = NameFormatConfig.objects.filter(service_name=self.service.name,
                                                        states__pk=self.user.profile.state.pk)

        if format_config.exists():
            return format_config[0]
        return None

    @cached_property
    def string_formatter(self):
        """
        Try to get the config format first
        Then the service default
        Before finally defaulting to global default
        :return: str
        """
        return getattr(self.formatter_config, 'format', self.default_formatter)

    @cached_property
    def default_formatter(self):
        return getattr(self.service, 'name_format', self.DEFAULT_FORMAT)

    @cached_property
    def _default_to_username(self):
        """
        Default to a users username if they have no main character.
        Default is True
        :return: bool
        """
        return getattr(self.formatter_config, 'default_to_username', True)



from django.urls import reverse


def get_admin_change_view_url(obj: object) -> str:
    """returns URL to admin change view for given object"""
    return reverse(
        'admin:{}_{}_change'.format(
            obj._meta.app_label, type(obj).__name__.lower()
        ),
        args=(obj.pk,)
    )

def get_admin_search_url(ModelClass: type) -> str:
    """returns URL to search URL for model of given object"""
    return '{}{}/'.format(
        reverse('admin:app_list', args=(ModelClass._meta.app_label,)),
        ModelClass.__name__.lower()
    )